/* Copyright 2004, 2005 Lei Zhang, Columbia University */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef _WIN32_WINNT
#define _WIN32_WINNT 0x0500
#endif
#include ".\tserver.h"
#include "TSProcMirrorCmd.h"
#include "MirrCmdProto.h"
#include "TSCommon.h"
#include "region.h"
#include <stdio.h>
#include "thincUtil.h"


/* copy from Ricardo's code */ 
#define T_BIT_FILL	0x8000	/* differentiate between a bitmap fill and just a bitmap */
#define T_ERROR -1
/***************/
//#define DBG_LOG_CMDS

extern ScreenStr screenStr;
extern ScreenPtr screenPtr;


typedef struct statPixmap_s
{
	int width;
	int height;
	int cmdCounter;
	int hash;
	int preCopyWidth;
	int preCopyHeight;
	int preCopyX;
	int preCopyY;
} StatPixmap;

StatPixmap statPixmapList[MAX_OFF_SURF+1];

#if 1
#define SMART_DISCARD
#endif 

#ifdef SMART_DISCARD

int preCounter = -1;
int preHash = -1;
int preWidth= -1;
int preHeight = -1;
int preCopyWidth= -1;
int preCopyHeight = -1;
int preCopyX= -1;
int preCopyY = -1;
#endif

void thincBlockHandler(thincScreenPtr pThinc);
BOOL useCopy();

/*  In windows, a pixmap is called a offscreen surface.
 *  We use both pixmap or offscreen surface in the code.
 */
extern PixmapPtr pixmapPtrTable[MAX_OFF_SURF+1];


/* try to merge r1 and r2. If it successes, store the merged rectangle in r1
 * and return true, otherwise return false */
static BOOL TryMergeRect(thinc_rectRegion* r1, thinc_rectRegion* r2)
{
	return FALSE;
	if (r1->height == r2->height && r1->y == r2->y)
	{
		if (r1->x + r1->width == r2->x)
		{
			/*
			   |r1|r2|
			 */
			r1->width += r2->width;
			return TRUE;
		} 
		else if (r2->x + r2->width == r1->x)
		{
			/*
			   |r2|r1|
			 */
			r1->width += r2->width;
			r1->x = r2->x;
			return TRUE;
		} 
		else
		{
			return FALSE;
		}
	} 
	else if (r1->width == r2->width && r1->x == r2->x)
	{
		if (r1->y + r1->height == r2->y)
		{
			/*
			    --
				r1
				--
				r2
				--
			*/
			r1->width += r2->width;
			return TRUE;
		}
		else if (r2->y + r2->height == r1->y)
		{
			/*
			    --
				r2
				--
				r1
				--
			*/
			r1->y = r2->y;
			r1->height += r2->y;
			return TRUE;
		} 
		else
		{
			return FALSE;
		}
	} 
	else
	{
		return FALSE;
	}
	
}

BOOL SendCmd(void* cmdData)
{
	void* pdata = cmdData;
	int i, res;

	MirrCmdHeader* header = (MirrCmdHeader*)pdata;

	if (header->type == T_CREATE_OFFSURF)
	{
		PixmapPtr pixmapPtr;
		thinc_createOffscreenSurf* pCreateOffScreen = 
			(thinc_createOffscreenSurf*)((char*)pdata + THINC_HEADERSIZE);
	
		
		if (header->dstId < MIN_OFF_SURF || header->dstId >= MAX_OFF_SURF)
		{
			/* The surface Id goes wrong*/
			__asm int 3; return FALSE;
		}
		
		if (pixmapPtrTable[header->dstId] != NULL)
		{
			/* The offscreen surface has exists */
			__asm int 3; return FALSE;             
		}
		
#ifdef SMART_DISCARD
		statPixmapList[header->dstId].width = pCreateOffScreen->width;
		statPixmapList[header->dstId].height = pCreateOffScreen->height;
		statPixmapList[header->dstId].hash = 0;
		statPixmapList[header->dstId].cmdCounter = 0;
		statPixmapList[header->dstId].preCopyWidth= -1;
		statPixmapList[header->dstId].preCopyHeight = -1;
		statPixmapList[header->dstId].preCopyX= -1;
		statPixmapList[header->dstId].preCopyY = -1;
#endif

		pixmapPtr = thincCreatePixmap(	screenPtr, 
										pCreateOffScreen->width, 
										pCreateOffScreen->height,
										pCreateOffScreen->depth);
		if (pixmapPtr == NULL)
		{
			/* Fail to create a fake PixmapRec structure*/
			__asm int 3; return FALSE;
		}                                            
		
		pixmapPtrTable[header->dstId] = pixmapPtr;

#ifdef INSTRU_CMD_DETAILS
		TSInstrument("CreateOffScreenSurface: id = %d, width = %d, height = %d, depth = %d",
			header->dstId,
			pCreateOffScreen->width, pCreateOffScreen->height, pCreateOffScreen->depth);
#endif 
		return TRUE;
	}

	else if (header->type == T_DESTROY_OFFSURF)
	{
		if (header->dstId < MIN_OFF_SURF || header->dstId >= MAX_OFF_SURF)
		{
			/* The surface Id goes wrong*/
			__asm int 3; return FALSE;
		}
		
		if (pixmapPtrTable[header->dstId] == NULL)
		{
			/* The offscreen surface doesn't exists */
			__asm int 3; return FALSE;             
		}

		res = thincDestroyPixmap(pixmapPtrTable[header->dstId]);

#ifdef INSTRU_CMD_DETAILS
		TSInstrument("DestroyOffScreenSurface: id = %d", header->dstId);
#endif
		pixmapPtrTable[header->dstId] = NULL;

		if (res != T_ERROR)
		{
			return TRUE;
		}
		else
		{
			__asm int 3; return FALSE;
		}

	}


	else if (header->type == T_FB_RUPDATE)
	{
		thinc_rawUpdate* rupdate = (thinc_rawUpdate*)((char*)pdata + THINC_HEADERSIZE);

		xRectangle* pXRect;
		RegionPtr pRegion;

#ifdef INSTRU_CMD_DETAILS
		TSInstrument("RAW_UPDATE: dst=%d, (x,y)=(%d,%d), (w,h)=(%d,%d)",
			header->dstId, rupdate->x, rupdate->y, rupdate->width, rupdate->height);
#endif

#ifdef SMART_DISCARD
		statPixmapList[header->dstId].hash <<= 1;
		statPixmapList[header->dstId].hash ^= header->type;
		statPixmapList[header->dstId].hash++;
		statPixmapList[header->dstId].cmdCounter++;
#endif

		pXRect = malloc(sizeof(xRectangle));
		if (pXRect == NULL)
		{
			__asm int 3; return FALSE;
		}

		pXRect->x = rupdate->x;
		pXRect->y = rupdate->y;
		pXRect->height = rupdate->height;
		pXRect->width = rupdate->width;

		pRegion = RECTS_TO_REGION(screenPtr, 1, pXRect, CT_NONE);
		if (pRegion == NULL)
		{
			__asm int 3; return FALSE;
		}

		if(header->dstId == FRAMEBUFF_ID) 
		{
			/* onscreen surface*/
			res = thincBufferNewCmdReg(screenPtr->pThinc, /*(WindowPtr)dst*/ NULL, 
				T_FB_RUPDATE, pRegion, 0, NULL); 
		}
		else
		{
			/* offscreen surface */
			if (header->dstId < MIN_OFF_SURF || header->dstId >= MAX_OFF_SURF)
			{
				/* The surface Id goes wrong*/
				__asm int 3; return FALSE;
			}

			if (pixmapPtrTable[header->dstId] == NULL)
			{
				/* The offscreen surface doesn't exists */
				__asm int 3; return FALSE;             
			}
			res = thincQueueCommand(screenPtr->pThinc, pixmapPtrTable[header->dstId], 
				T_FB_RUPDATE, pRegion, 0, NULL); 
		}

		free(pXRect); //??????
		REGION_DESTROY(screenPtr, pRegion);

		return TRUE;
	} 
	else if (header->type == T_FB_SFILL)
	{
		thinc_fbSFill* sfill = (thinc_fbSFill*)((char*)pdata + THINC_HEADERSIZE);
		thinc_rectRegion* pTRect = (thinc_rectRegion*)((char*)pdata + THINC_HEADERSIZE + T_FB_SFILL_SIZE);
		xRectangle* pXRectList;
		RegionPtr pRegion;
		int recNum;
		pXRectList = malloc(sizeof(xRectangle)* sfill->numrects);

#ifdef INSTRU_CMD_DETAILS
		TSInstrument("SOLID_FILL  : dst=%d, color = %x", header->dstId, sfill->pixel);
#endif

#ifdef SMART_DISCARD
		statPixmapList[header->dstId].hash <<= 1;
		statPixmapList[header->dstId].hash ^= header->type;
		statPixmapList[header->dstId].hash++;
		statPixmapList[header->dstId].cmdCounter++;
#endif

		if (pXRectList == NULL)
		{
			__asm int 3; return FALSE;
		}

		for (i = 0, recNum = 0; i < sfill->numrects; i++)
		{
			BOOL canMerge = FALSE;
			if (i < sfill->numrects - 1) 
			{
				canMerge = TryMergeRect(&pTRect[i], &pTRect[i+1]);
			}
			if (!canMerge)
			{
				pXRectList[recNum].x = pTRect[i].x;
				pXRectList[recNum].y = pTRect[i].y;
				pXRectList[recNum].height = pTRect[i].height;
				pXRectList[recNum].width = pTRect[i].width;
				recNum++;
			}
#ifdef INSTRU_CMD_DETAILS
			TSInstrument("  SOLID_FILL: dst=%d, color = %x, (x,y)=(%d,%d), (w,h)=(%d,%d) ", 
				header->dstId, sfill->pixel, 
				pTRect[i].x, pTRect[i].y, pTRect[i].width, pTRect[i].height);
#endif
		}

		pRegion = RECTS_TO_REGION(screenPtr, recNum, pXRectList, CT_NONE);

		if (pRegion == NULL)
		{
			__asm int 3; return FALSE;
		}

		if (header->dstId == FRAMEBUFF_ID)
		{
			/* On screen surface */
			res = thincBufferNewCmdReg(screenPtr->pThinc, /*(WindowPtr)dst*/ NULL, T_FB_SFILL, 
				pRegion, sfill->pixel);
		}
		else
		{
			/* offscreen surface */
			if (header->dstId < MIN_OFF_SURF || header->dstId >= MAX_OFF_SURF)
			{
				/* The surface Id goes wrong*/
				__asm int 3; return FALSE;
			}

			if (pixmapPtrTable[header->dstId] == NULL)
			{
				/* The offscreen surface doesn't exists */
				__asm int 3; return FALSE;             
			}

			res = thincQueueCommand(screenPtr->pThinc, pixmapPtrTable[header->dstId], 
				T_FB_SFILL, pRegion, sfill->pixel);
		}       

		free(pXRectList); //??????
		REGION_DESTROY(screenPtr, pRegion);

		return TRUE;
		//if (Send_fbSFill(0, sfill->pixel, sfill->numrects,
		//	(thinc_rectRegion*)((char*)pdata + THINC_HEADERSIZE + T_FB_SFILL_SIZE)) < 0 )
		//{
		//	__asm int 3; return FALSE;
		//}
	} 
	else if (header->type == T_FB_PFILL)
	{
		thinc_fbPFill*    pfill      = (thinc_fbPFill*)((char*)pdata + THINC_HEADERSIZE);
		int               pixmapSize = pfill->width * pfill->height * fb.bpp;
		thinc_rectRegion* pTRect     = (thinc_rectRegion*)((char*)pdata
			                             + THINC_HEADERSIZE + T_FB_PFILL_SIZE + pixmapSize);
		char*			  pixmap	 = (char*)pdata + THINC_HEADERSIZE + T_FB_PFILL_SIZE;

		xRectangle*       pXRectList;
		RegionPtr         pRegion;
		int               recNum;

		pXRectList = malloc(sizeof(xRectangle)* pfill->numrects);

#ifdef INSTRU_CMD_DETAILS
		TSInstrument("PIXMAP_FILL  : dst=%d, (x,y)=(%d,%d), (w,h)=(%d,%d)", 
			header->dstId, pfill->x, pfill->y, pfill->width, pfill->height);
#endif

#ifdef SMART_DISCARD
		statPixmapList[header->dstId].hash <<= 1;
		statPixmapList[header->dstId].hash ^= header->type;
		statPixmapList[header->dstId].hash++;
		statPixmapList[header->dstId].cmdCounter++;
#endif

		if (pXRectList == NULL)
		{
			__asm int 3; return FALSE;
		}

		for (i = 0, recNum = 0; i < pfill->numrects; i++)
		{
			pXRectList[recNum].x = pTRect[i].x;
			pXRectList[recNum].y = pTRect[i].y;
			pXRectList[recNum].height = pTRect[i].height;
			pXRectList[recNum].width = pTRect[i].width;
			recNum++;
#ifdef INSTRU_CMD_DETAILS
			TSInstrument("  PIXMAP_FILL: dst=%d, (x,y)=(%d,%d), (w,h)=(%d,%d) ", 
				header->dstId, pTRect[i].x, pTRect[i].y, pTRect[i].width, pTRect[i].height);
#endif
		}

		pRegion = RECTS_TO_REGION(screenPtr, recNum, pXRectList, CT_NONE);

		if (pRegion == NULL)
		{
			__asm int 3; return FALSE;
		}

		if (header->dstId == FRAMEBUFF_ID)
		{
			/* On screen surface */
			res = thincBufferNewCmdReg(screenPtr->pThinc, /*(WindowPtr)dst*/ NULL, T_FB_PFILL, 
				pRegion, pixmap, pfill->x, pfill->y, pfill->width, pfill->height);
		}
		else
		{
			/* offscreen surface */
			if (header->dstId < MIN_OFF_SURF || header->dstId >= MAX_OFF_SURF)
			{
				/* The surface Id goes wrong*/
				__asm int 3; return FALSE;
			}

			if (pixmapPtrTable[header->dstId] == NULL)
			{
				/* The offscreen surface doesn't exists */
				__asm int 3; return FALSE;             
			}

			res = thincQueueCommand(screenPtr->pThinc, pixmapPtrTable[header->dstId], 
				T_FB_PFILL, pRegion, pixmap, pfill->x, pfill->y, pfill->width, pfill->height);
		}       

		free(pXRectList); //??????
		REGION_DESTROY(screenPtr, pRegion);

		return TRUE;

	} 
	else if (header->type == T_FB_GLYPH)
	{
		//printf("StartNew(): Obtained a SFILL command\n");
		//return TRUE;
		thinc_fbGlyph*	glyph = (thinc_fbGlyph*)((char*)pdata + THINC_HEADERSIZE);
		void*			bitmap = (char*)pdata + THINC_HEADERSIZE + T_FB_GLYPH_SIZE;
  		xRectangle*		pXRectList;
		RegionPtr		pRegion;
		int				numRects;
		int				i;

#ifdef INSTRU_CMD_DETAILS
		TSInstrument("GLYPH  : dst=%d, color=%x, (x,y)=(%d,%d), (w,h)=(%d,%d), numRects=%d", 
			header->dstId, glyph->color, glyph->x, glyph->y, 
			glyph->width, glyph->height, glyph->numrects);
#endif

#ifdef SMART_DISCARD
		statPixmapList[header->dstId].hash <<= 1;
		statPixmapList[header->dstId].hash ^= header->type;
		statPixmapList[header->dstId].hash++;
		statPixmapList[header->dstId].cmdCounter++;
#endif

		if (glyph->numrects == 0)
		{
			pXRectList = malloc(sizeof(xRectangle));
			if (pXRectList == NULL)
			{
				__asm int 3; return FALSE;
			}
			pXRectList->x = glyph->x;
			pXRectList->y = glyph->y;
			pXRectList->width = glyph->width;
			pXRectList->height = glyph->height;
			numRects = 1;
		}
		else
		{
			thinc_rectRegion* pTRect;
			pTRect = (thinc_rectRegion*)((char*)bitmap + (glyph->width+7)/8*glyph->height);
			pXRectList = malloc(sizeof(xRectangle)* glyph->numrects);
			numRects = glyph->numrects;

			if (pXRectList == NULL)
			{
				__asm int 3; return FALSE;
			}
			for (i = 0; i < glyph->numrects; i++)
			{
				pXRectList[i].x = pTRect[i].x;
				pXRectList[i].y = pTRect[i].y;
				pXRectList[i].height = pTRect[i].height;
				pXRectList[i].width = pTRect[i].width;
#ifdef INSTRU_CMD_DETAILS
				TSInstrument("  GLYPH: (x,y)=(%d,%d), (w,h)=(%d,%d)", 
					pTRect[i].x, pTRect[i].y, pTRect[i].width, pTRect[i].height);
#endif
			}
		}                                 

		pRegion = RECTS_TO_REGION(screenPtr, numRects, pXRectList, CT_NONE);

		if (pRegion == NULL)
		{
			__asm int 3; return FALSE;
		}

		if(header->dstId == FRAMEBUFF_ID)
		{
			res = thincBufferNewCmdReg(screenPtr->pThinc, /*(WindowPtr)dst*/NULL, T_FB_GLYPH, pRegion, 
				bitmap, glyph->x, glyph->y, glyph->width, glyph->height, glyph->color, T_BIT_FILL);
		}
		else 
		{
			if (header->dstId < MIN_OFF_SURF || header->dstId >= MAX_OFF_SURF)
			{
				/* The surface Id goes wrong*/
				__asm int 3; return FALSE;
			}

			if (pixmapPtrTable[header->dstId] == NULL)
			{
				/* The offscreen surface doesn't exists */
				__asm int 3; return FALSE;             
			}

			res = thincQueueCommand(screenPtr->pThinc, pixmapPtrTable[header->dstId], T_FB_GLYPH,
				pRegion, bitmap, glyph->x, glyph->y, glyph->width, glyph->height, glyph->color,
				T_BIT_FILL);
		}

		free(pXRectList); //??????
		REGION_DESTROY(screenPtr, pRegion);

		return TRUE;

		//if (Send_fbGlyph(0, glyph->x, glyph->y, glyph->width, glyph->height, glyph->color,
		//	glyph->numrects, bitmap, (thinc_rectRegion*)rectlist) < 0 )
		//{
		//	__asm int 3; return FALSE;
		//}
	} 
	else if (header->type == T_FB_COPY)
	{
		thinc_fbCopy* copy = (thinc_fbCopy*)((char*)pdata + THINC_HEADERSIZE);
		//printf("StartNew(): Obtained a COPY command\n");

		if (header->dstId < FRAMEBUFF_ID || header->dstId >= MAX_OFF_SURF)
		{
			/* The surface Id goes wrong*/
			__asm int 3; 
			return FALSE;
		}

		if (header->srcId < FRAMEBUFF_ID || header->srcId >= MAX_OFF_SURF)
		{
			/* The surface Id goes wrong*/
			__asm int 3; 
			return FALSE;
		}

		if (header->srcId != FRAMEBUFF_ID && pixmapPtrTable[header->srcId] == NULL)
		{
			/* The offscreen surface doesn't exists */
			__asm int 3; 
			return FALSE;             
		}

		if (header->dstId != FRAMEBUFF_ID && pixmapPtrTable[header->dstId] == NULL)
		{
			/* The offscreen surface doesn't exists */
			__asm int 3; 
			return FALSE;             
		}

#ifdef INSTRU_CMD_DETAILS
		TSInstrument("COPY: %d-->%d, (x,y):(%d,%d)-->(%d,%d), (w,h)=(%d,%d)",
			header->srcId, header->dstId, copy->sx, copy->sy, copy->dx, copy->dy,
			copy->width, copy->height);
#endif
		if (header->dstId == FRAMEBUFF_ID && header->srcId == FRAMEBUFF_ID)
		{
			RegionRec copyReg;
			/* onscreen --> onscreen
			 *
			 * Flush all previous commands*/
			thincBlockHandler(screenPtr->pThinc);

			/* From thincCopyArea_onScreen() in thincGC.c, it seems that
			 * a bunch of clipping should be done. But I don't think we 
			 * need to do them here.
			 */
			coords_to_reg(screenPtr, &copyReg, copy->sx, copy->sy, copy->width, copy->height);

			if(useCopy()) 
			{
				res = thincBufferNewCmdReg(screenPtr->pThinc, /*WindowPtr*/ NULL, T_FB_COPY, 
					&copyReg, (copy->dx - copy->sx), (copy->dy - copy->sy));
				UnblockDriverCopying();
			} else 
			{
				res = thincBufferNewCmdReg(screenPtr->pThinc, /*WindowPtr*/ NULL, T_FB_RUPDATE, &copyReg,
					0, NULL);
			}
			
			if (res == T_ERROR)
				__asm int 3; return FALSE;
			else
				return TRUE;

		} else if (header->dstId == FRAMEBUFF_ID && header->srcId != FRAMEBUFF_ID)
		{
			/* offscreen --> onscreen */
			RegionRec dstReg, fbReg;

#ifdef SMART_DISCARD
			if (statPixmapList[header->srcId].cmdCounter == preCounter &&
				statPixmapList[header->srcId].hash == preHash &&
				statPixmapList[header->srcId].width == preWidth &&
				statPixmapList[header->srcId].height == preHeight &&
				copy->sx == statPixmapList[header->dstId].preCopyX &&
				copy->sy == statPixmapList[header->dstId].preCopyY &&
				copy->width == statPixmapList[header->dstId].preCopyWidth &&
				copy->height == statPixmapList[header->dstId].preCopyHeight &&
				statPixmapList[header->srcId].cmdCounter > 100)
			{
				//preCounter = statPixmapList[header->srcId].cmdCounter;
				//preHash = statPixmapList[header->srcId].cmdCounter;
				//preWidth = statPixmapList[header->srcId].width;
				//preHeight = statPixmapList[header->srcId].height;
				//__asm int 3;
				return TRUE;
			}
			else
			{
				preCounter = statPixmapList[header->srcId].cmdCounter;
				preHash = statPixmapList[header->srcId].hash;
				preWidth = statPixmapList[header->srcId].width;
				preHeight = statPixmapList[header->srcId].height;
				statPixmapList[header->dstId].preCopyX = copy->sx;
				statPixmapList[header->dstId].preCopyY = copy->sy;
				statPixmapList[header->dstId].preCopyWidth = copy->width;
				statPixmapList[header->dstId].preCopyHeight = copy->height;
			}
#endif

			coords_to_reg(screenPtr, &dstReg, copy->dx, copy->dy, copy->width, copy->height);
			coords_to_reg(screenPtr, &fbReg, 0, 0, fb.width, fb.height);

			REGION_INTERSECT(screenPtr, &dstReg, &fbReg, &dstReg);

			if (!REGION_NOTEMPTY(screenPtr, &dstReg))
			{
				/* It should not happen. */
				__asm int 3;
				__asm int 3; return FALSE;
			}

			if(useCopy()) 
			{
				res = thincExecPixmapQueue(	screenPtr->pThinc,
					/*WindowPtr*/ NULL,
					pixmapPtrTable[header->srcId],
					copy->sx,
					copy->sy,
					copy->width,
					copy->height,
					copy->dx,
					copy->dy,
					&dstReg);
			} 
			else 
			{
				res = thincBufferNewCmdReg(screenPtr->pThinc, /*WindowPtr*/ NULL, T_FB_RUPDATE, &dstReg,
					0, NULL);
			}

			if (res != T_ERROR)
			{
				return TRUE;
			}
			else
			{	
				__asm int 3; return FALSE;
			}


		} else if (header->dstId != FRAMEBUFF_ID && header->srcId == FRAMEBUFF_ID)
		{
			/* onscreen  --> offscreen */
			RegionRec dstReg;
			coords_to_reg(screenPtr, &dstReg, copy->dx, copy->dy, copy->width, copy->height);
			res = thincQueueCommand(screenPtr->pThinc, pixmapPtrTable[header->dstId], 
				T_FB_RUPDATE, &dstReg, 0, NULL); 
			if (res != T_ERROR)
			{
				return TRUE;
			}
			else
			{
				__asm int 3; return FALSE;
			}
		} else if (header->dstId != FRAMEBUFF_ID && header->srcId != FRAMEBUFF_ID)
		{
			/* offscreen --> offscreen */
			//if (useCopy())
			//{
			//	__asm int 3;
			//}

			res = thincCopyPixmapQueue(	screenPtr->pThinc, 
										pixmapPtrTable[header->srcId], 
										pixmapPtrTable[header->dstId],
										copy->sx, copy->sy,
										copy->width, copy->height,
										copy->dx, copy->dy);
			if (res != T_ERROR)
			{
				return TRUE;
			}
			else
			{
				__asm int 3; return FALSE;
			}
		}

		//if (Send_fbCopy(0, copy->sx, copy->sy, copy->dx, copy->dy,
		//	copy->width, copy->height) < 0)
		//{
		//	return FALSE;
		//}				
	}
	else if (header->type == T_CURSOR_CHANGE)
	{
		thinc_cursorChange* cursorchange = (thinc_cursorChange*)((char*)pdata + THINC_HEADERSIZE);
		cursor.source = (char*)cursorchange + T_CURSOR_CHANGE_SIZE;
		cursor.mask = (char*)cursor.source + cursor.cursorSize;
		if (cursor.cursorSize <= 0)
		{
			printf("StartNew(): previous cursor size = 0, the client doesn't know the cursor size yet.\n");
			__asm int 3; return FALSE;
		}
		
		CursorWin32ToX((BYTE*)cursor.source, (BYTE*)cursor.mask, cursor.cursorSize);

		thincLoadCursorImage(&screenStr, (unsigned char*)cursor.source);

#ifdef INSTRU_CMD_DETAILS
		TSInstrument("CURSOR_CHANGE: cursor (w,h) = (%d,%d)", 
			cursor.cursorWidth, cursor.cursorHeight);
#endif
		//Send_cursorChange(&cursor);
	}

	else if (header->type = T_TIME_STAMP)
	{
		thinc_timeStamp* timeStamp = (thinc_timeStamp*)((char*)pdata + THINC_HEADERSIZE);
		SYSTEMTIME st;
		
		DriverSysTimeToUserSysTime((PTIME_FIELDS)&timeStamp->time, &st);

		switch (timeStamp->stampType)
		{
			case TIME_STAMP_ORIG:
				TSInstrument("Time stamp: "\
					"%.4d/%.2d/%.2d - %.2d:%.2d:%.2d:%.3d  :: Original Command: %s",
					st.wYear, st.wMonth, st.wDay, 
					st.wHour, st.wMinute, st.wSecond, st.wMilliseconds,
					timeStamp->record);
				break;
			case TIME_STAMP_THINC:
				TSInstrument("Time stamp: "\
					"%.4d/%.2d/%.2d - %.2d:%.2d:%.2d:%.3d  :: THINC Command: %s",
					st.wYear, st.wMonth, st.wDay, 
					st.wHour, st.wMinute, st.wSecond, st.wMilliseconds,
					timeStamp->record);
					break;
			case TIME_STAMP_BEFORE_WAIT:
				TSInstrument("Time stamp: "\
					"%.4d/%.2d/%.2d - %.2d:%.2d:%.2d:%.3d  :: Before Waiting: %s",
					st.wYear, st.wMonth, st.wDay, 
					st.wHour, st.wMinute, st.wSecond, st.wMilliseconds,
					timeStamp->record);
					break;
			case TIME_STAMP_AFTER_WAIT:
				TSInstrument("Time stamp: "\
					"%.4d/%.2d/%.2d - %.2d:%.2d:%.2d:%.3d  :: After Waiting: %s",
					st.wYear, st.wMonth, st.wDay, 
					st.wHour, st.wMinute, st.wSecond, st.wMilliseconds,
					timeStamp->record);
					break;
			case TIME_STAMP_DRIVER_FUNCTION:
				TSInstrument("Time stamp: "\
					"%.4d/%.2d/%.2d - %.2d:%.2d:%.2d:%.3d  :: Driver function: %s",
					st.wYear, st.wMonth, st.wDay, 
					st.wHour, st.wMinute, st.wSecond, st.wMilliseconds,
					timeStamp->record);
					break;
			case TIME_STAMP_AFTER_DRAWING:
				TSInstrument("Time stamp: "\
					"%.4d/%.2d/%.2d - %.2d:%.2d:%.2d:%.3d  :: After Drawing: %s",
					st.wYear, st.wMonth, st.wDay, 
					st.wHour, st.wMinute, st.wSecond, st.wMilliseconds,
					timeStamp->record);
					break;
		}
		return FALSE;
	}else {
		TSInstrument("Unknown Commands");
		//printf("StartNew(): Obtained an unkown command\n");
		__asm int 3; return FALSE;
	}
	return TRUE;
}

void DriverSysTimeToUserSysTime(PTIME_FIELDS pDrvST, PSYSTEMTIME pST)
{
	pST->wDay = pDrvST->Day;
	pST->wDayOfWeek = pDrvST->Weekday;
	pST->wHour = pDrvST->Hour;
	pST->wMilliseconds = pDrvST->Milliseconds;
	pST->wMinute = pDrvST->Minute;
	pST->wMonth = pDrvST->Month;
	pST->wSecond = pDrvST->Second;
	pST->wYear = pDrvST->Year;
}