/* Copyright 2004, 2005 Lei Zhang, Columbia University */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef _WIN32_WINNT
#define _WIN32_WINNT 0x0500
#endif
#include ".\tserver.h"
#include <stdio.h>
#include <tchar.h>
#include <string.h>
#include <winsock2.h>
#include "compact.h"
#include "TSProcMirrorCmd.h"
#include "thinc.h"
#include "TSEvents.h"
#include "TSDebug.h"

LPSTR driverName = "Microsoft Mirror Driver";
LPSTR eventName = "THINCSharedEvent";
const char server_version[3] = {0, 1, 1};

int loopTimeoutInterval = 3000;

LPSTR dispCode[7] = {
   "Change Successful",
   "Must Restart",
   "Bad Flags",
   "Bad Parameters",
   "Failed",
   "Bad Mode",
   "Not Updated"};

LPSTR enNewCmd				= "THINCNewCommand";
LPSTR enBlockOnDraw			= "THINCBlockOnDraw";
LPSTR enRestart				= "THINCRestart";
LPSTR enStartBlocking		= "THINCStartBlocking";
LPSTR enBlockOnCopy			= "THINCBlockOnCopy";
LPSTR enStartBlockingCopy	= "THINCStartBlockingCopy";


#define INPUT_MOUSE_MOTION    1
#define INPUT_MOUSE_CLICK     2
#define INPUT_KEYBOARD_EVENT  3

#define DRV_RECHK_INTERVAL 3000	    /* 100-nanoseconds */

#define MY_RECHK_INTERVAL 2			/* milliseconds */

//#define POLLING

typedef struct thincInputEvent_s {
	int type;
	union {
		struct {
			int x;
			int y;
		} mouseMotion;
		struct {
			int button;
			int isdown;
		} mouseClick;
		struct {
			int key;
			int isdown;
		} keyboardEvent;
	};
} thincInputEvent;
SOCKET eventSocket;

char inputBuf[102400];	//100KB

ScreenStr screenStr;
ScreenPtr screenPtr = &screenStr;

PixmapPtr pixmapPtrTable[MAX_OFF_SURF];
		
static void thincMouseMoveReal( int x, int y);
static void thincMouseButtonReal(int button, int isdown);
static void thincKeyboardStrokeReal(int vkey, int isdown);


static DWORD CALLBACK InputEventProc(LPVOID parameter)
{
	SOCKET ListenSocket;
	SOCKET AcceptSocket;
	struct sockaddr_in service;
	thincInputEvent nextInputEvent;
	//----------------------
	// Create a SOCKET for listening for
	// incoming connection requests.
	ListenSocket = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
	if (ListenSocket == INVALID_SOCKET) {
		__asm int 3;
		WSACleanup();
		return 0;
	}

	//----------------------
	// The sockaddr_in structure specifies the address family,
	// IP address, and port for the socket that is being bound.
	service.sin_family = AF_INET;
	service.sin_addr.s_addr = inet_addr("127.0.0.1");
	service.sin_port = htons(19999);	//internal communication

	if (bind( ListenSocket, 
		(SOCKADDR*) &service, 
		sizeof(service)) == SOCKET_ERROR) {
			__asm int 3;
			closesocket(ListenSocket);
			return 0;
	}

	//----------------------
	// Listen for incoming connection requests.
	// on the created socket
	if (listen( ListenSocket, 1 ) == SOCKET_ERROR)
		printf("Error listening on socket.\n");

	//----------------------
	// Create a SOCKET for accepting incoming requests.
	printf("Waiting for client to connect...\n");

	//----------------------
	// Accept the connection.
	while(1) {
		AcceptSocket = SOCKET_ERROR;
		while( AcceptSocket == SOCKET_ERROR ) {
			AcceptSocket = accept( ListenSocket, NULL, NULL );
		}
		printf("Client connected.\n");
		ListenSocket = AcceptSocket;
		break;
	}
	

	for (;;)
	{
		int res = recv(AcceptSocket, &nextInputEvent, sizeof(thincInputEvent), 0);
		if (res == 0)
		{
			// socket closed
			__asm int 3;
			return 0;
		}
		if (res == SOCKET_ERROR)
		{
			// socket error;
			res = WSAGetLastError();
			__asm int 3;		
			return 0;
		}
		if (nextInputEvent.type == INPUT_MOUSE_MOTION)
		{
			thincMouseMoveReal( nextInputEvent.mouseMotion.x,
								nextInputEvent.mouseMotion.y);
		}
		else if (nextInputEvent.type == INPUT_MOUSE_CLICK)
		{
			thincMouseButtonReal(nextInputEvent.mouseClick.button,
								 nextInputEvent.mouseClick.isdown);
		}
		else if (nextInputEvent.type == INPUT_KEYBOARD_EVENT)
		{
			thincKeyboardStrokeReal(nextInputEvent.keyboardEvent.key,
									nextInputEvent.keyboardEvent.isdown);
		}
	}

	return 0;
}

void TServer(void)
{
	hKeyProfileMirror = (HKEY)0;
	hKeyDevice = (HKEY)0;
}

void TServerTearDown(void)
{
}

// Get the HDC of the mirror driver. Return TRUE on success
BOOL getMirrDC(void)
{
	    
	DEVMODE devmode;
	BOOL change;

    FillMemory(&devmode, sizeof(DEVMODE), 0);

    devmode.dmSize = sizeof(DEVMODE);
    devmode.dmDriverExtra = 0;

    change = EnumDisplaySettings(NULL,
                                      ENUM_CURRENT_SETTINGS,
                                      &devmode);

    devmode.dmFields = DM_BITSPERPEL |
                       DM_PELSWIDTH | 
                       DM_PELSHEIGHT;
  
	if (change) 
    {
		// query all display devices in the system until we hit
        // our favourate mirrored driver, then extract the device name string
        // of the format '\\.\DISPLAY#'
       
        DISPLAY_DEVICE dispDevice;
        LPSTR deviceName = NULL;
        INT devNum = 0;
        BOOL result;
        CHAR deviceNum[MAX_PATH];
        LPSTR deviceSub;
        DWORD one = 1;
		INT code;

        FillMemory(&dispDevice, sizeof(DISPLAY_DEVICE), 0);
       
        dispDevice.cb = sizeof(DISPLAY_DEVICE);
       

        devmode.dmDeviceName[0] = '\0';


        while (result = EnumDisplayDevices(NULL,
                                  devNum,
                                  &dispDevice,
                                  0))
        {
			if (strcmp(&dispDevice.DeviceString[0], driverName) == 0)
				break;
			devNum++;
        }
       
        if (!result)
        {
			printf("No '%s' found.\n", driverName);
			return FALSE;
		}

        printf("DevNum:%d\nName:%s\nString:%s\nID:%s\nKey:%s\n\n",
               devNum,
               &dispDevice.DeviceName[0],
               &dispDevice.DeviceString[0],
               &dispDevice.DeviceID[0],
               &dispDevice.DeviceKey[0]);


        // Simply extract 'DEVICE#' from registry key.  This will depend
        // on how many mirrored devices your driver has and which ones
        // you intend to use.

        _strupr(&dispDevice.DeviceKey[0]);

        deviceSub = strstr(&dispDevice.DeviceKey[0],
                           "\\DEVICE");

        if (!deviceSub) 
            strcpy(&deviceNum[0], "DEVICE0");
        else
            strcpy(&deviceNum[0], ++deviceSub);
        
        // Add 'Attach.ToDesktop' setting.
        //

        hKeyProfileMirror = (HKEY)0;
        if (RegCreateKey(HKEY_LOCAL_MACHINE,
                        _T("SYSTEM\\CurrentControlSet\\Hardware Profiles\\Current\\System\\CurrentControlSet\\Services\\mirror"),
                         &hKeyProfileMirror) != ERROR_SUCCESS)
        {
           printf("Can't access registry.\n");
           return 0;
        }

		hKeyDevice = (HKEY)0;
        if (RegCreateKey(hKeyProfileMirror,
                         _T(&deviceNum[0]),
                         &hKeyDevice) != ERROR_SUCCESS)
        {
			printf("Can't access DEVICE# hardware profiles key.\n");
			return FALSE;
        }

        if (RegSetValueEx(hKeyDevice,
                          _T("Attach.ToDesktop"),
                          0,
                          REG_DWORD,
                          (unsigned char *)&one,
                          4) != ERROR_SUCCESS)
        {
			printf("Can't set Attach.ToDesktop to 0x1\n");
			return FALSE;
		}

        strcpy((LPSTR)&devmode.dmDeviceName[0], "mirror");
        deviceName = (LPSTR)&dispDevice.DeviceName[0];

        // add 'Default.*' settings to the registry under above hKeyProfile\mirror\device
        code =
        ChangeDisplaySettingsEx(deviceName,
                                &devmode, 
                                NULL,
                                CDS_UPDATEREGISTRY,
                                NULL
                                );
    
        printf("Update Register on device mode: %s\n", GetDispCode(code));

        code = ChangeDisplaySettingsEx(deviceName,
                                &devmode, 
                                NULL,
                                0,
                                NULL
                                );
   
        printf("Raw dynamic mode change on device mode: %s\n", GetDispCode(code));

		mirrHdc = CreateDC("DISPLAY",
							deviceName,
							NULL,
							NULL);
	}
    else
    {
        printf("Can't get display settings.\n");
		return FALSE;
    }
		
	return TRUE;
}

LPSTR GetDispCode(INT code)
{
   static char tmp[MAX_PATH];
   switch (code) {
   
   case DISP_CHANGE_SUCCESSFUL: return dispCode[0];
   
   case DISP_CHANGE_RESTART: return dispCode[1];
   
   case DISP_CHANGE_BADFLAGS: return dispCode[2];
   
   case DISP_CHANGE_BADPARAM: return dispCode[3];
   
   case DISP_CHANGE_FAILED: return dispCode[4];
   
   case DISP_CHANGE_BADMODE: return dispCode[5];
   
   case DISP_CHANGE_NOTUPDATED: return dispCode[6];
   
   default:
      sprintf(&tmp[0],"Unknown code: %08x\n", code);
      return (LPTSTR)&tmp[0];
   
   }
   
   return NULL;   // can't happen
}

BOOL releaseMirr(void)
{
	DWORD zero = 0;
	DeleteDC(mirrHdc);

	// detach mirror driver from the desktop
	if (RegSetValueEx(hKeyDevice,
		_T("Attach.ToDesktop"),
		0,
		REG_DWORD,
		(unsigned char *)&zero,
		4) != ERROR_SUCCESS)
	{
		printf("Can't set Attach.ToDesktop to 0x0\n");
		return FALSE;
	}

	RegCloseKey(hKeyProfileMirror);
	RegCloseKey(hKeyDevice);

	return TRUE;
}

BOOL Init(void)
{
	if (!getMirrDC())
	{
		printf("Fail to create a device context for the mirror display driver\n");
		return FALSE;
	}
	return TRUE;
}
DWORD WINAPI StartNew(LPVOID parameter)
{
	int res;
	int i;
	char InBuf[50];
	char OutBuf[2000];

	int		cmdSize;
	void*	cmd;

	PThincCmdPtr pCmdPtr;
	//LableAndEvent lae;
	HANDLE hThread;
	DWORD dwThreadId;

	PTIME_FIELDS pTimeStamp;
	SYSTEMTIME st;

	HANDLE hEventClientInput;

	freopen("c:\\serverlog.txt", "a", stdout);

	memset(InBuf, 0, sizeof(InBuf));
	memset(OutBuf, 0, sizeof(OutBuf));
	//
	//  All events are created as auto-reseted and non-signaled 
	// 

	if ((hEventClientInput = CreateEvent(NULL, FALSE, FALSE, NULL)) == NULL)
	{
		printf(" StartNew(): Could not open hEventClientInput\n");
		printf(" StartNew(): GetLastError() = %d\n", GetLastError());
		return FALSE;
	}

	if ((hEventNewCmd = CreateEvent(NULL, FALSE, FALSE, enNewCmd)) == NULL)
	{
		printf(" StartNew(): Could not open hEventNewCmd\n");
		printf(" StartNew(): GetLastError() = %d\n", GetLastError());
		return FALSE;
	}

	if ((hEventBlockOnDraw = CreateEvent(NULL, FALSE, FALSE, enBlockOnDraw)) == NULL)
	{
		printf(" StartNew(): Could not open hEventBlockOnDraw\n");
		printf(" StartNew(): GetLastError() = %d\n", GetLastError());
		return FALSE;
	}

	if ((hEventStartBlocking = CreateEvent(NULL, FALSE, FALSE, enStartBlocking)) == NULL)
	{
		printf(" StartNew(): Could not open hEventStartBlocking\n");
		printf(" StartNew(): GetLastError() = %d\n", GetLastError());
		return FALSE;
	}

	if ((hEventRestart = CreateEvent(NULL, FALSE, FALSE, enRestart)) == NULL)
	{
		printf(" StartNew(): Could not open hEventRestart\n");
		printf(" StartNew(): GetLastError() = %d\n", GetLastError());
		return FALSE;
	}

	if ((hEventBlockOnCopy = CreateEvent(NULL, FALSE, FALSE, enBlockOnCopy)) == NULL)
	{
		printf(" StartNew(): Could not open hEventBlockOnCopy\n");
		printf(" StartNew(): GetLastError() = %d\n", GetLastError());
		return FALSE;
	}

	if ((hEventStartBlockingCopy = CreateEvent(NULL, FALSE, FALSE, enStartBlockingCopy)) == NULL)
	{
		printf(" StartNew(): Could not open hEventStartBlockingCopy\n");
		printf(" StartNew(): GetLastError() = %d\n", GetLastError());
		return FALSE;
	}

	hThread = CreateThread(NULL, 0, InputEventProc, NULL, 0, &dwThreadId);

	do {
		struct sockaddr_in clientService; 
		eventSocket = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
		if (eventSocket == INVALID_SOCKET) {
			__asm int 3;
			return 0;
		}

		clientService.sin_family = AF_INET;
		clientService.sin_addr.s_addr = inet_addr( "127.0.0.1" );
		clientService.sin_port = htons( 19999 );

		if ( connect( eventSocket, (SOCKADDR*) &clientService, sizeof(clientService) ) == SOCKET_ERROR) {
			__asm int 3;
			return 0;
		}

	} while(FALSE);


	/* Add the new command event as the first event in the event list*/
	AddEvent(hEventClientInput);

	/* Intialize the offscreen pixmap pointer table */
	for (i = 0; i < MAX_OFF_SURF; i++)
	{
		pixmapPtrTable[i] = NULL;
	}


	printf("Before ExtEscape...\n");

	((PParaIn)InBuf)->recheckInterval = DRV_RECHK_INTERVAL;

	if (res = ExtEscape(mirrHdc, 
						GESC_SERVER_START,
						sizeof(InBuf),
						&InBuf[0],
						sizeof(OutBuf),
						&OutBuf[0]) == GESC_SERVER_START) 
	{
		POINT cursorPos;
		PMapToUserMode mappingInfo = (PMapToUserMode)(&OutBuf[0]);
		mirror_cursor* mc;
		thinc_createOffscreenSurf* pCreateFB;

		for (i = 0; i < mappingInfo->len; i++)
		{
			if (mappingInfo->MapToUserList[i].flags == MM_FG_ON_SURF)
			{
				fb.fb = mappingInfo->MapToUserList[i].userVirtualAddress;
			}
			else if (mappingInfo->MapToUserList[i].flags == MM_FG_CMD_QUE)
			{
				pCmdPtrQueue = (PThincCmdPtrQueue)(mappingInfo->MapToUserList[i].userVirtualAddress);
			}
			else if (mappingInfo->MapToUserList[i].flags == MM_FG_CMD_BLK)
			{
				pCmdBlock = mappingInfo->MapToUserList[i].userVirtualAddress;
			}
			else if (mappingInfo->MapToUserList[i].flags == MM_FG_SIGNAL)
			{
				pSignals = (PThincSignals)(mappingInfo->MapToUserList[i].userVirtualAddress);
			}
		}

		mc = (mirror_cursor*)(OutBuf + sizeof(MapToUserMode));

		//printf(" mc is pointing to: %d\n", (int)mc);
		printf("cursor size: %d, height: %d, width %d, xHot: %d, yHot: %d\n",
			mc->cursorSize, mc->cursorHeight, mc->cursorWidth, mc->xHot, mc->yHot);

		//mc->cursorSize = -1;
		cursor.cursorSize = mc->cursorSize / 2;
		cursor.cursorBG = 0xffffff;
		cursor.cursorFG = 0;
		cursor.cursorHeight = mc->cursorHeight;
		cursor.cursorWidth = mc->cursorWidth;
		cursor.xHot = mc->xHot;
		cursor.yHot = mc->yHot;
		if (cursor.cursorSize > 0) 
		{
			cursor.source = (char*)mc + sizeof(mirror_cursor);
			cursor.mask = (char*)cursor.source + cursor.cursorSize;
		}

		GetCursorPos(&cursorPos);

		//
		// Read the first command "creating framebuffer". Extract the framebuffer's info
		// 

		printf("StartNew(): Start reading framebuffer information\n");
		printf("StartNew(): headIndex = %d, TailIndex=%d\n", pCmdPtrQueue->cmdHeadIndex,
			pCmdPtrQueue->cmdTailIndex);

		pCmdPtr = popCmd(pCmdPtrQueue);

		if (pCmdPtr == NULL)
		{
			printf("StartNew(): NULL command pointer\n");
			return FALSE;
		}

		printf("StartNew(): headIndex = %d, TailIndex=%d\n", pCmdPtrQueue->cmdHeadIndex,
			pCmdPtrQueue->cmdTailIndex);
		printf("StartNew(): Get the commands\n");

		cmdSize = *(int*)((char*)(pCmdBlock) + pCmdPtr->offset);
		printf("StartNew(): Get the commands. Size = %d\n", cmdSize);
		cmd = ((char*)(pCmdBlock)) + pCmdPtr->offset + sizeof(int) + sizeof(TIME_FIELDS);
		pCreateFB = (thinc_createOffscreenSurf*)((char*)cmd + THINC_HEADERSIZE + 8); // FIXME: header_size

		fb.bpp = pCreateFB->bpp;		
		fb.depth = 24;// FIXME: pCreateFB->depth;
		fb.height = pCreateFB->height;
		fb.width = pCreateFB->width;
		fb.stride = pCreateFB->stride;

		ThincInit(&screenStr, fb.fb, fb.width, fb.height, fb.depth, fb.bpp * 8);// FIXEME: bpp
		ThincCursorInit(&screenStr, &cursor);

		CursorWin32ToX(cursor.source, cursor.mask, cursor.cursorSize);

		thincLoadCursorImage(screenPtr, cursor.source);
		printf("StartNew(): Framebuffer info: width = %d, height = %d, depth = %d, bpp = %d, stride = %d\n",
			fb.width, fb.height, fb.depth, fb.bpp, fb.stride);


	} 
	else 
	{
		printf("StartNew(): Error in getting the frame buffer\n");
		return FALSE;
	}

	printf("-------------------------------------------------------------------------\n");
	freopen("serverlog.txt", "a", stdout);
#ifdef _DEBUG
	thincStatsInit(screenPtr->pThinc);
#endif
#ifdef POLLING
	for (;;)
	{
		int currentCmdCount, res;
#if 0
		currentCmdCount = QUEUELENGTH(pCmdPtrQueue);

		if (currentCmdCount == 0)
		{
			res = WaitForSingleObject(hEventClientInput, 1);

			if (res == WAIT_OBJECT_0)
			{
				// There are user inputs
				thincWakeupHandler(screenStr.pThinc);
			}
		}
		else
		{
			res = WAIT_TIMEOUT;
		}
#else
		res = WaitForSingleObject(hEventClientInput, 1);

		if (res == WAIT_OBJECT_0)
		{
			// There are user inputs
			thincWakeupHandler(screenStr.pThinc);
		}
#endif
		//printf("StartNew():: pEventSocketList->num = %d\n", pEventSocketList->num);

		pSignals->bReqNewCommandSignal = TRUE;
		currentCmdCount = QUEUELENGTH(pCmdPtrQueue);
		TSInstrument("Command Number: %d", currentCmdCount);
		for (i = 0; i < currentCmdCount; i++)
		{
			pCmdPtr = popCmd(pCmdPtrQueue);
			if (pCmdPtr != NULL)
			{
				cmdSize = *(int*)((char*)(pCmdBlock) + pCmdPtr->offset);
				// FIXME: can we compute the pointers without the sizeof macro?
				pTimeStamp = ((char*)(pCmdBlock)) + pCmdPtr->offset + sizeof(int);
				cmd = ((char*)(pCmdBlock)) + pCmdPtr->offset + sizeof(int) + sizeof(TIME_FIELDS);
				if (!SendCmd(cmd))
				{
					//
					// It returns false when it process a time stamp command
					//
				}	
				else
				{
#ifdef _DEBUG
					DriverSysTimeToUserSysTime(pTimeStamp, &st);
					TSInstrument(" Command Generated at: %.4d/%.2d/%.2d - %.2d:%.2d:%.2d:%.3d",
						st.wYear, st.wMonth, st.wDay, 
						st.wHour, st.wMinute, st.wSecond, st.wMilliseconds);
#endif
				}
			}
		}

		//printf("StartNew():: --> thincWakeupHandler\n");
		if (res != WAIT_OBJECT_0)
		{
			// we call it flush the remaining data in the buffer to clients
			thincWakeupHandler(screenStr.pThinc);
		}
		//printf("StartNew():: <-- thincWakeupHandler\n");
		
		//BlockDriverDrawing();
		//printf("StartNew():: --> thincBlockHandler\n");
		thincBlockHandler(screenStr.pThinc);
		//printf("StartNew():: <-- thincBlockHandler\n");
		//UnblockDriverDrawing();
	}
#else /* POLLING */
	TSInstrumentEnd();
	TSInstrumentStart();
	for (;;)
	{
		int currentCmdCount;

		WaitForSingleObject(hEventNewCmd, 10);
		pSignals->bReqNewCommandSignal = TRUE;
		currentCmdCount = QUEUELENGTH(pCmdPtrQueue);
		TSInstrument("Command Number: %d", currentCmdCount);
		for (i = 0; i < currentCmdCount; i++)
		{
			pCmdPtr = popCmd(pCmdPtrQueue);
			if (pCmdPtr != NULL)
			{
				
				cmdSize = *(int*)((char*)(pCmdBlock) + pCmdPtr->offset);
				pTimeStamp = ((char*)(pCmdBlock)) + pCmdPtr->offset + sizeof(int);
				cmd = ((char*)(pCmdBlock)) + pCmdPtr->offset + sizeof(int) + sizeof(TIME_FIELDS);
#ifdef _DEBUG
				DriverSysTimeToUserSysTime(pTimeStamp, &st);
				TSInstrument(" Command Generated at: %.4d/%.2d/%.2d - %.2d:%.2d:%.2d:%.3d",
					st.wYear, st.wMonth, st.wDay, 
					st.wHour, st.wMinute, st.wSecond, st.wMilliseconds);
#endif
				if (!SendCmd(cmd))
				{
					printf("StartNew(): Error in SendCmd. Command size = %d\n", cmdSize);
					//return FALSE;
				}		
			}
		}

		thincWakeupHandler(screenStr.pThinc);
		
		//BlockDriverDrawing();
		thincBlockHandler(screenStr.pThinc);
		//UnblockDriverDrawing();
	}
#endif /* POLLING */
	return TRUE;
}

BOOL End(void)
{
	releaseMirr();
	return TRUE;
}

BOOL useCopy()
{
	return ((thincScreenPtr)(screenPtr->pThinc))->useCopy;
}

static void thincMouseMoveReal(int x, int y)
{
	mouse_event(MOUSEEVENTF_ABSOLUTE | MOUSEEVENTF_MOVE, 
				x * 65535 / fb.width, 
				y * 65535 / fb.height,
				0,
				0);
}


void thincMouseButtonReal(int button, int isdown)
{
	if (isdown)
	{
		//
		// Button down
		// 
		if ((button & 0x7f) == LBUTTON)
		{
			mouse_event(MOUSEEVENTF_LEFTDOWN, 0, 0, 0, 0);
		}
		else if ((button & 0x7f) == RBUTTON)
		{
			mouse_event(MOUSEEVENTF_RIGHTDOWN, 0, 0, 0, 0);
		}
		else if ((button & 0x7f) == MBUTTON)
		{
			mouse_event(MOUSEEVENTF_MIDDLEDOWN, 0, 0, 0, 0);
		}

	}		
	else
	{
		//
		// Button up
		//
		if ((button & 0x7f) == LBUTTON)
		{
			mouse_event(MOUSEEVENTF_LEFTUP, 0, 0, 0, 0);
		}
		else if ((button & 0x7f) == RBUTTON)
		{
			mouse_event(MOUSEEVENTF_RIGHTUP, 0, 0, 0, 0);
		}
		else if ((button & 0x7f) == MBUTTON)
		{
			mouse_event(MOUSEEVENTF_MIDDLEUP, 0, 0, 0, 0);
		}
	}

}

static void thincKeyboardStrokeReal(int vkey, int isdown)
{
	if (isdown)
	{
		keybd_event(vkey, 0, 0, 0);
	}
	else
	{
		keybd_event(vkey, 0, KEYEVENTF_KEYUP, 0);
	}
}

void thincMouseMove(int x, int y)
{
	thincInputEvent nextInputEvent;
	nextInputEvent.type = INPUT_MOUSE_MOTION;
	nextInputEvent.mouseMotion.x = x;
	nextInputEvent.mouseMotion.y = y;
	send(eventSocket, (char*)&nextInputEvent, sizeof(thincInputEvent), 0);
}

void thincMouseButton(int button, int isdown)
{
	thincInputEvent nextInputEvent;
	nextInputEvent.type = INPUT_MOUSE_CLICK;
	nextInputEvent.mouseClick.button = button;
	nextInputEvent.mouseClick.isdown = isdown;
	
	if (((button & 0x7f) == LBUTTON) && !isdown)
	{
		// Add a separator to the instrument log
		TSInstrumentEnd();
		TSInstrumentStart();
	}

	send(eventSocket, (char*)&nextInputEvent, sizeof(thincInputEvent), 0);
}

void thincKeyboardStroke(int vkey, int isdown)
{
	thincInputEvent nextInputEvent;
	nextInputEvent.type = INPUT_KEYBOARD_EVENT;
	nextInputEvent.keyboardEvent.key = vkey;
	nextInputEvent.keyboardEvent.isdown = isdown;
#ifdef _DEBUG
	//reset the statistic for each keyboard input. 
	//It is not that nice. but I don't have time 
	thincStatsPrint(screenPtr->pThinc);
	thincStatsReset(screenPtr->pThinc);
#endif
	send(eventSocket, (char*)&nextInputEvent, sizeof(thincInputEvent), 0);
}


void CursorWin32ToX(BYTE* source, BYTE* mask, int cursorSize)
{
	int i, j;
	for (i = 0; i < cursorSize; i++)
	{
		BYTE sb1 = source[i];
		BYTE sb2 = 0;
		BYTE mb1 = mask[i];
		BYTE mb2 = 0;
		BYTE bm1 = 1;
		BYTE bm2 = 0x80;
		for (j = 0; j < 8; j++)
		{
			if (sb1 & bm1) 
			{
				//(1, *)
				sb2 |= bm2;
			}

			if (mb1 & bm1)
			{
				//(*, 1)
				mb2 |= bm2;
			}
			else
			{
				// 
				//(0,0)-->(1,1) Make it compatible with Linux 
				//
				if (!(sb1 & bm1))
				{
					sb2 |= bm2;
					mb2 |= bm2;
				}
			}
			bm1 <<= 1;
			bm2 >>= 1;
		}
		source[i] = sb2;
		mask[i] = mb2;
	}
}
#if 1
void BlockDriverDrawing()
{
	/* Note: InterlockedCompareExchange returns the original value */
	int counter = 0;
	int waitTime = MY_RECHK_INTERVAL;
	TSInstrument("Try to grabbing framebuffer lock");
	while (InterlockedCompareExchange(&(pSignals->bStatusDrawing), TRUE, FALSE))
	{
		TSInstrument("Try to grabbing framebuffer lock. Go to sleep");
//		Sleep(waitTime); 
		Sleep(18);
		TSInstrument("Try to grabbing framebuffer lock. Waitup");
		if (counter++ % 50 == 0) 
		{
			//waitTime += MY_RECHK_INTERVAL * 10;
			//if (counter > 300) 
			//	break;
		}
	}
}

void UnblockDriverDrawing()
{
	pSignals->bStatusDrawing = FALSE;
}
#else
void BlockDriverDrawing()
{
	/* Note: InterlockedCompareExchange returns the original value */
	int counter = 0;
	int waitTime = MY_RECHK_INTERVAL;
	if (InterlockedCompareExchange(&(pSignals->bStatusDrawing), TRUE, FALSE))
	{
		
	}
}

void UnblockDriverDrawing()
{
	pSignals->bStatusDrawing = FALSE;
}

#endif
void UnblockDriverCopying()
{
	// 
	// The driver is blocked before copying, unblock it
	// 
	pSignals->bStatusBlockBeforeCopy = FALSE;
	SetEvent(hEventBlockOnCopy);
}