/*
 *
 * Header for data compression
 *
 * Author:
 * Ricardo Baratto
 * Network Computing Lab - Columbia University
 *
 */

/* set tabstop = 8 */

/* the idea is to create a compression API we can use to wrap around the
 * actual compression engine we're using. It's modeled after zlib since
 * that's what i know right now. The API tries to have a streaming
 * model (again, similar to zlib's) */
#ifdef T_WIN32
#ifndef T_USEPNG
#include "zlib.h"
#else
#include "png.h"
#endif
#else /* T_WIN32 */
#ifndef T_USEPNG
#include "zlib.h"
#else
#include "png.h"
#endif
#endif /* T_WIN32 */

typedef struct thincEncode_S {
	/* compressed buffer and its size */
	unsigned char *			zbuff;
	int				zbuff_size;

	/* compression level */
	int				level;
#ifndef T_USEPNG
	/* zlib specific objects */
	z_stream *			zstream;
#endif
#ifdef T_USEPNG
	/* png specific objects */
	png_structp			png_ptr;
	png_infop			info_ptr;
#endif
	unsigned char *			data;
	unsigned char *			zcurr;
	int				curr_size;
} thincEncodeRec, *thincEncodePtr;

int thinc_compressInit(thincScreenPtr pThinc, int compress_level);
void thinc_compressEnd(thincScreenPtr pThinc);
int thinc_compressPrepare(thincScreenPtr pThinc);
int thinc_compress(thincScreenPtr pThinc);
int thinc_compressFlush(thincScreenPtr pThinc);
unsigned char * thinc_compressGetBuffer(thincScreenPtr pThinc);
int thinc_compressMaxSize(thincScreenPtr pThinc);

int thinc_compressSetParams(thincScreenPtr pThinc, int params, ...);
#define T_ZPARAM_IN	0x01
#define T_ZPARAM_OUT	0x02
#define T_ZPARAM_W	0x04
#define T_ZPARAM_H	0x08
#define T_ZPARAM_BPP	0x10	/* FIXME: NOT USED */

/* how much worse-case overhead will zlib have. From zlib.h:
 * "the destination buffer [must be] at least 0.1% larger than [source buffer] 
 * plus 12 bytes"
 */
#define T_Z_OVERHEAD(size) ((((double)(size))*0.01) + 12)

/* what's the definite minimum size a compressed update will have. 12 bytes
 * is actually less than that if we're using zlib (see above) */
#define T_Z_MINSIZE		T_FB_ZDATA_SIZE + T_CL_MINSIZE
