/*
 *
 * The THINC application sharing protocol
 *
 * Authors:
 * Ricardo Baratto
 * Network Computing Lab - Columbia University
 */

/* set tabstop = 8 */

/* based on 
 * 'Protocols for Application and Desktop Sharing'
 * http://www1.cs.columbia.edu/~lennox/draft-lennox-avt-app-sharing-00.html
 */

/* header. It's not the same as defined by the draft document, as we want to
 * support multiple applications with multiple windows on the same connection.
 * The draft defines the header to be:
 * x_off: 32bit
 * y_off: 32bit
 * win_id: 16bit
 * mbz: 8bit
 * pt: 8bit (payload type)
 *
 * x and y depend on the message type so we don't put them in the header.
 * win_id is still 16 bits.
 * pt: defined in thinc_header as the message type (see thincProto.h)
 */
/* XXX HACK ALERT XXX */
/* to have minimum disruption, when application sharing is being used
 * this message will immediately follow the thinc_header. When app
 * sharing is not being used, then there's no change from before.
 * In protocol v2, we have a way to negotiate this at connection time.
 * In the old protocol, i wish you the best of luck.
 */
typedef struct thinc_appHeader_s {
	int app_id:8;
	int mbz:8;
	int win_id:16;
} thinc_appHeader;

#define T_APPHEADER_SIZE	4

/* Description of an application */
typedef struct thinc_appDesc_s {
	int	id		:8;	/* id of app. used for all comm */
	int	name_len	:8;	/* the length of the name 
					   (without final \0) */
	int	win_count	:16;	/* how many windows the app has */
	/* followed by name_len chars which are the name of the app */
} thinc_appDesc;

#define T_APP_DESC_SIZE		4

/* the application id of the whole desktop. The rest of the ids are generated
 * by the server */
#define T_APP_ID_DESKTOP	0
