/* Copyright 2003, 2004, 2005 Ricardo Baratto */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/

/*
 * 
 * RENDER related functions for THINC
 *
 * Author:
 * Ricardo Baratto
 * Network Computing Lab - Columbia University
 */

#include "thinc.h"

#ifdef T_USERENDER

#if 0
#define WRITE_GLYPH_TOFILE_FUN
#endif

#if 0
#define WRITE_GLYPH_TOFILE
#endif

#ifdef WRITE_GLYPH_TOFILE_FUN
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <netinet/in.h>

static void thincDumpGlyph(unsigned char *data, int width, int height);
#endif	/* WRITE_GLYPH_TOFILE_FUN */

#define thincPict_pre(pPict, func) \
	ScreenPtr pScreen = pPict->pDrawable->pScreen; \
	PictureScreenPtr ps = GetPictureScreen(pScreen); \
	thincScreenPtr pThinc = THINCPtr(pScreen); \
	thinc_unwrap(pThinc, ps, func);

#define thincPict_post(ps, pThinc, func) \
	thinc_wrap(pThinc, ps, func, thinc##func);

static void thincGlyphs1Bpp(thincScreenPtr pThinc, PicturePtr pSrc, 
	PicturePtr pDst, INT16 xSrc, INT16 ySrc, 
	int nlists, GlyphListPtr lists, GlyphPtr *glyphs);
static int thincGetPictureColor(thincScreenPtr pThinc,
	PicturePtr pic, unsigned long *color);

int composite_stay_quiet = 0;

#define img_coords(pImage, xc, yc) 		\
	(xc) += (pImage)->pDrawable->x;		\
	(yc) += (pImage)->pDrawable->y;

/* For RENDER. Based on fbPictureInit. Assumes standard PictureScreenRec
 * has been initialized (driver should call fbPictureInit() or 
 * miPictureInit()) */
int thincPictureInit(ScreenPtr pScreen)
{
	t_dbg_name(func)
	PictureScreenPtr ps;
	thincScreenPtr pThinc = THINCPtr(pScreen);

	thincDbg(T_DBG_DRAW, ("%s\n", func));

	ps = GetPictureScreen(pScreen);

	thinc_wrap(pThinc, ps, CreatePicture, thincCreatePicture);
	thinc_wrap(pThinc, ps, DestroyPicture, thincDestroyPicture);
	thinc_wrap(pThinc, ps, ChangePictureClip, thincChangePictureClip);
	thinc_wrap(pThinc, ps, DestroyPictureClip, thincDestroyPictureClip);
	thinc_wrap(pThinc, ps, ChangePicture, thincChangePicture);
	thinc_wrap(pThinc, ps, ValidatePicture, thincValidatePicture);
	thinc_wrap(pThinc, ps, InitIndexed, thincInitIndexed);
	thinc_wrap(pThinc, ps, CloseIndexed, thincCloseIndexed);
	thinc_wrap(pThinc, ps, UpdateIndexed, thincUpdateIndexed);

	/* The actual rendering functions */
	thinc_wrap(pThinc, ps, Composite, thincComposite);
	thinc_wrap(pThinc, ps, Glyphs, thincGlyphs);
	thinc_wrap(pThinc, ps, CompositeRects, thincCompositeRects);
	thinc_wrap(pThinc, ps, Trapezoids, thincTrapezoids);
	thinc_wrap(pThinc, ps, Triangles, thincTriangles);
	thinc_wrap(pThinc, ps, TriStrip, thincTriStrip);
	thinc_wrap(pThinc, ps, TriFan, thincTriFan);
	thinc_wrap(pThinc, ps, RasterizeTrapezoid, thincRasterizeTrapezoid);

	/* print some very useful information */
	thincDbg(T_DBG_DRAW, ("%s: these are the picture formats:\n"
		"32bpp\n"
		"\tPICT_a8r8g8b8: %#x\n"
		"\tPICT_x8r8g8b8: %#x\n"
		"\tPICT_a8b8g8r8: %#x\n"
		"\tPICT_x8b8g8r8: %#x\n"
		"24bpp\n"
		"\tPICT_r8g8b8: %#x\n"
		"\tPICT_b8g8r8: %#x\n"
		"16bpp\n"
		"\tPICT_r5g6b5: %#x\n"
		"\tPICT_b5g6r5: %#x\n"
		"\tPICT_a1r5g5b5: %#x\n"
		"\tPICT_x1r5g5b5: %#x\n"
		"\tPICT_a1b5g5r5: %#x\n"
		"\tPICT_x1b5g5r5: %#x\n"
		"\tPICT_a4r4g4b4: %#x\n"
		"\tPICT_x4r4g4b4: %#x\n"
		"\tPICT_a4b4g4r4: %#x\n"
		"\tPICT_x4b4g4r4: %#x\n"
		"8bpp\n"
		"\tPICT_a8: %#x\n"
		"\tPICT_r3g3b2: %#x\n"
		"\tPICT_b2g3r3: %#x\n"
		"\tPICT_a2r2g2b2: %#x\n"
		"\tPICT_a2b2g2r2: %#x\n"
		"\tPICT_c8: %#x\n"
		"\tPICT_g8: %#x\n",
		func,
		/* 32bpp */
		PICT_a8r8g8b8, PICT_x8r8g8b8, PICT_a8b8g8r8, PICT_x8b8g8r8,
		/* 24bpp */
		PICT_r8g8b8, PICT_b8g8r8,
		/* 16bpp */
		PICT_r5g6b5, PICT_b5g6r5, PICT_a1r5g5b5, PICT_x1r5g5b5,
		PICT_a1b5g5r5, PICT_x1b5g5r5, PICT_a4r4g4b4, PICT_x4r4g4b4,
		PICT_a4b4g4r4, PICT_x4b4g4r4,
		/* 8bpp */
		PICT_a8, PICT_r3g3b2, PICT_b2g3r3, PICT_a2r2g2b2,
		PICT_a2b2g2r2, PICT_c8, PICT_g8
	));

	thincDbg(T_DBG_DRAW, ("%s: the rest of the picture formats:\n"
		"4bpp\n"
		"\tPICT_a4: %#x\n"
		"\tPICT_r1g2b1: %#x\n"
		"\tPICT_b1g2r1: %#x\n"
		"\tPICT_a1r1g1b1: %#x\n"
		"\tPICT_a1b1g1r1: %#x\n"
		"\tPICT_c4: %#x\n"
		"\tPICT_g4: %#x\n"
		"1bpp\n"
		"\tPICT_a1: %#x\n"
		"\tPICT_g1: %#x\n",
		func,
		/* 4bpp */
		PICT_a4, PICT_r1g2b1, PICT_b1g2r1, PICT_a1r1g1b1,
		PICT_a1b1g1r1, PICT_c4, PICT_g4,
		/* 1bpp */
		PICT_a1, PICT_g1
	));


	thincDbg(T_DBG_DRAW, ("%s: finished\n", func));
	return T_OK;
}

/* unwrap all functions. Call from CloseScreen() */
void thincPictureUninit(ScreenPtr pScreen)
{
	t_dbg_name(func)
	PictureScreenPtr ps;
	thincScreenPtr pThinc = THINCPtr(pScreen);

	thincDbg(T_DBG_DRAW, ("%s\n", func));
	
	ps = GetPictureScreen(pScreen);

	thinc_unwrap(pThinc, ps, CreatePicture);
	thinc_unwrap(pThinc, ps, DestroyPicture);
	thinc_unwrap(pThinc, ps, ChangePictureClip);
	thinc_unwrap(pThinc, ps, DestroyPictureClip);
	thinc_unwrap(pThinc, ps, ChangePicture);
	thinc_unwrap(pThinc, ps, ValidatePicture);
	thinc_unwrap(pThinc, ps, InitIndexed);
	thinc_unwrap(pThinc, ps, CloseIndexed);
	thinc_unwrap(pThinc, ps, UpdateIndexed);
	thinc_unwrap(pThinc, ps, Composite);
	thinc_unwrap(pThinc, ps, Glyphs);
	thinc_unwrap(pThinc, ps, CompositeRects);

	thincDbg(T_DBG_DRAW, ("%s: finished\n", func));
}

int thincCreatePicture (PicturePtr pPict)
{
	t_dbg_name(func)
	int res;
	thincPict_pre(pPict, CreatePicture);

	thincDbg(T_DBG_DRAW, ("%s: picture format is %#lx. repeat? %d\n", func,
		pPict->format, pPict->repeat));
	thincDrawable_print(pPict->pDrawable, "");

	res = ps->CreatePicture(pPict);
	thincPict_post(ps, pThinc, CreatePicture);

	thincDbg(T_DBG_DRAW, ("%s: finished\n", func));

	return res;
}

void thincDestroyPicture(PicturePtr pPict)
{
	t_dbg_name(func)
	thincPict_pre(pPict, DestroyPicture);

	thincDbg(T_DBG_DRAW, ("%s\n", func));
	thincDrawable_print(pPict->pDrawable, "");

	ps->DestroyPicture(pPict);
	thincPict_post(ps, pThinc, DestroyPicture);
	
	thincDbg(T_DBG_DRAW, ("%s: finished\n", func));
}

int thincChangePictureClip(PicturePtr pPict, int clipType, pointer value, 
	int n)
{
	t_dbg_name(func)
	int res;
	thincPict_pre(pPict, ChangePictureClip);

	thincDbg(T_DBG_DRAW, ("%s\n", func));
	thincDrawable_print(pPict->pDrawable, "");

	res = ps->ChangePictureClip(pPict, clipType, value, n);
	thincPict_post(ps, pThinc, ChangePictureClip);

	thincDbg(T_DBG_DRAW, ("%s: finished\n", func));

	return res;
}

void thincDestroyPictureClip(PicturePtr pPict)
{
	t_dbg_name(func)
	thincPict_pre(pPict, DestroyPictureClip);

	thincDbg(T_DBG_DRAW, ("%s\n", func));
	thincDrawable_print(pPict->pDrawable, "");

	ps->DestroyPictureClip(pPict);
	thincPict_post(ps, pThinc, DestroyPictureClip);

	thincDbg(T_DBG_DRAW, ("%s: finished\n", func));
}
    
void thincChangePicture(PicturePtr pPict, Mask mask)
{
	t_dbg_name(func)
	thincPict_pre(pPict, ChangePicture);

	thincDbg(T_DBG_DRAW, ("%s: mask %#lx\n", func, mask));
	thincDrawable_print(pPict->pDrawable, "");

	ps->ChangePicture(pPict, mask);
	thincPict_post(ps, pThinc, ChangePicture);

	thincDbg(T_DBG_DRAW, ("%s: finished\n", func));
}

void thincValidatePicture(PicturePtr pPict, Mask mask)
{
	t_dbg_name(func)
	thincPict_pre(pPict, ValidatePicture);

	thincDbg(T_DBG_DRAW, ("%s: mask: %#lx\n", func, mask));
	thincDrawable_print(pPict->pDrawable, "");

	ps->ValidatePicture(pPict, mask);
	thincPict_post(ps, pThinc, ValidatePicture);

	thincDbg(T_DBG_DRAW, ("%s: finished\n", func));
}

Bool thincInitIndexed(ScreenPtr pScreen, PictFormatPtr pFormat)
{
	t_dbg_name(func)
	Bool res;
	PictureScreenPtr ps = GetPictureScreen(pScreen);
	thincScreenPtr pThinc = THINCPtr(pScreen);
	thinc_unwrap(pThinc, ps, InitIndexed);

	thincDbg(T_DBG_DRAW, ("%s\n", func));

	res = ps->InitIndexed(pScreen, pFormat);
	thincPict_post(ps, pThinc, InitIndexed);

	thincDbg(T_DBG_DRAW, ("%s: finished\n", func));

	return res;
}

void thincCloseIndexed(ScreenPtr pScreen, PictFormatPtr pFormat)
{
	t_dbg_name(func)
	PictureScreenPtr ps = GetPictureScreen(pScreen);
	thincScreenPtr pThinc = THINCPtr(pScreen);
	thinc_unwrap(pThinc, ps, CloseIndexed);

	thincDbg(T_DBG_DRAW, ("%s\n", func));

	ps->CloseIndexed(pScreen, pFormat);
	thincPict_post(ps, pThinc, CloseIndexed);

	thincDbg(T_DBG_DRAW, ("%s: finished\n", func));
}

void thincUpdateIndexed(ScreenPtr pScreen, PictFormatPtr pFormat, int ndef, 
	xColorItem *pdef)
{
	t_dbg_name(func)
	PictureScreenPtr ps = GetPictureScreen(pScreen);
	thincScreenPtr pThinc = THINCPtr(pScreen);
	thinc_unwrap(pThinc, ps, UpdateIndexed);

	thincDbg(T_DBG_DRAW, ("%s\n", func));

	ps->UpdateIndexed(pScreen, pFormat, ndef, pdef);
	thincPict_post(ps, pThinc, UpdateIndexed);

	thincDbg(T_DBG_DRAW, ("%s: finished\n", func));
}

/*******************************************************************/

/* the way composite works is combining pSrc and pMask, and drawing to 
 * pDst. at a basic level, pMask basically tells us the shape (has the alpha 
 * values), while pSrc tells us the content (has pixel values).
 * and we combine those and blast them into pDst */
void thincComposite(CARD8 op, PicturePtr pSrc, PicturePtr pMask, 
	PicturePtr pDst, INT16 xSrc, INT16 ySrc, INT16 xMask, INT16 yMask, 
	INT16 xDst, INT16 yDst, CARD16 width, CARD16 height)
{
	t_dbg_name(func)
	int isOff;
	RegionRec reg;
	DrawablePtr dst;
	int err;
	thincPict_pre(pSrc, Composite);

	thincDbg(T_DBG_DRAW, ("%s: op %d. params: "
		"xSrc %d, ySrc %d, xMask %d, yMask %d, xDst %d, yDst %d "
		"width %d , height %d\n", func, op,
		xSrc, ySrc, xMask, yMask, xDst, yDst, width, height));
	thincDbg(T_DBG_DRAW, ("%s: formats: src %#lx, dst %#lx\n", func,
		pSrc->format, pDst->format));
	thincDrawable_print(pSrc->pDrawable, "src: ");
	thincDrawable_print(pDst->pDrawable, "dst: ");

	dst = pDst->pDrawable;

	/* exec the op */
	ps->Composite(op, pSrc, pMask, pDst, xSrc, ySrc, xMask,
		yMask, xDst, yDst, width, height);

	/* XXX: HACK? Probably should store this in the image private? */
	if(composite_stay_quiet) {
		thincDbg(T_DBG_DRAW, ("%s: somebody's got it, nothing to do\n",
			func));
		goto end;
	}
	thincDbg(T_DBG_DRAW, ("%s: nobody's got it, better get to it\n", 
		func));

	really_print_reg(T_DBG_DRAW, pThinc->pScreen, pDst->pCompositeClip,
		"dst composite clip");

	if(pSrc) { img_coords(pSrc, xSrc, ySrc); }
	if(pDst) { img_coords(pDst, xDst, yDst); }
	if(pMask) { img_coords(pMask, xMask, yMask); }

	REGION_INIT(pThinc->pScreen, &reg, &thincNullBox, 1);
	err = miComputeCompositeRegion(&reg, pSrc, pMask, pDst, xSrc, ySrc,
		xMask, yMask, xDst, yDst, width, height);
	if(!err) {
		thincDbg(T_DBG_DRAW, ("%s: Composite region is empty or "
			"something failed\n", func));
		goto out;
	}
	if(region_is_empty(&reg)) {
		thincDbg(T_DBG_DRAW, ("%s: Draw region is empty, finished\n",
			func));
		goto out;
	}

	isOff = is_offscreen(dst);
	if(isOff) {
		err = thincQueueCommand(pThinc, (PixmapPtr)pDst->pDrawable,
			T_FB_RUPDATE, &reg, 0, NULL);
	} else {
		err = thincBufferNewCmdReg(pThinc, (WindowPtr)pDst->pDrawable,
			T_FB_RUPDATE, &reg, 0, NULL);
	}
	if(err == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't queue RAW "
			"update for composite\n", func));
		goto out;
	}

	thincDbg(T_DBG_DRAW, ("%s: finished\n", func));
out:
	/* it may have been UNINIT-ed by miComputeCompositeRegion() but
	 * hopefully it's smart enough to figure that out */
	REGION_UNINIT(pThinc->pScreen, &reg);
end:
	thincPict_post(ps, pThinc, Composite);
}

void thincCompositeRects (CARD8 op, PicturePtr pDst, xRenderColor *color, 
	int nRect, xRectangle *rects)
{
	t_dbg_name(func)
	int i;
	thincPict_pre(pDst, CompositeRects);

	thincDbg(T_DBG_DRAW, ("%s: %d\n", func, op));
	thincDrawable_print(pDst->pDrawable, "");

	for(i=0; i<nRect; ++i) {
		thincDbg(T_DBG_DRAW, ("%s: rect[%d]: %d %d %d %d\n", func,
			i, rects[i].x, rects[i].y, rects[i].width,
			rects[i].height));
	}

	ps->CompositeRects(op, pDst, color, nRect, rects);
	thincPict_post(ps, pThinc, CompositeRects);

	thincDbg(T_DBG_DRAW, ("%s: finished\n", func));
}

void thincGlyphs (CARD8 op, PicturePtr pSrc, PicturePtr pDst, 
	PictFormatPtr maskFormat, INT16 xSrc, INT16 ySrc, int nlists, 
	GlyphListPtr lists, GlyphPtr *glyphs)
{
	t_dbg_name(func)
	void (*glyph_func)(thincScreenPtr, PicturePtr, PicturePtr, INT16, 
		INT16, int, GlyphListPtr, GlyphPtr *) = NULL;
	
	thincPict_pre(pSrc, Glyphs);

	thincDbg(T_DBG_DRAW, ("%s: op %d, nlists %d. parameters: "
		"xSrc %d, ySrc %d\n", func, op, nlists, xSrc, ySrc));
	thincDrawable_print(pSrc->pDrawable, "src: ");
	thincDrawable_print(pDst->pDrawable, "dst: ");

#if 0	/* XXX Debug XXX */
	thincDbg(T_DBG_DRAW, ("%s: Src picture information:\t"
		"id %lu, format %#lx, repeat %d\n", func,
		pSrc->id, pSrc->format, pSrc->repeat));
	thincDbg(T_DBG_DRAW, ("%s: Dst picture information:\t"
		"id %lu, format %#lx, repeat %d\n", func,
		pDst->id, pDst->format, pDst->repeat));

	if(maskFormat) {
		thincDbg(T_DBG_DRAW, ("%s: have mask format id %ld, "
			"format %#lx, type %#x, depth %d\n", func,
			maskFormat->id, maskFormat->format,
			maskFormat->type, maskFormat->depth));
	} 
	/* XXX Debug XXX */
#endif

	/* tell Composite() that we got it all under control up here */
	if(!maskFormat) {
		/* FIXME: This happens when openoffice is doing antialiased
		 * text. tried looking at the pSrc format and it comes out
		 * as 24bpp.
		 * And if we disable antialiased text then it starts using 
		 * FillRect. these people are insane. openoffice blows. */ 
		composite_stay_quiet = 0;
		goto glyphs;
	} 
	switch(maskFormat->depth) {
		case 1:
			thincDbg(T_DBG_DRAW, ("%s: Doing non-antialiased "
				"text\n", func));
			glyph_func = thincGlyphs1Bpp;
			composite_stay_quiet = 1;
			break;
		case 8:
			/* TODO: This is antialiased text on a solid
			 * background */
#if 0
			T_BUG("Implement depth 8!");
#endif
			glyph_func = NULL;
			composite_stay_quiet = 0;
			break;
		/* FIXME: Which others? (24 and 32 haven been spotted) */
		default:
#if 0
			T_BUG_ARG("Implement support for depth %d", 
				maskFormat->depth);
#endif
			glyph_func = NULL;
			composite_stay_quiet = 0;
			break;
	}

glyphs:
	/* get this over with before we mess things up */
	ps->Glyphs(op, pSrc, pDst, maskFormat, xSrc, ySrc, nlists,
		lists, glyphs);
	thincPict_post(ps, pThinc, Glyphs);

	if(!composite_stay_quiet) {
		/* nothing else to do */
		thincDbg(T_DBG_DRAW, ("%s: finished\n", func));
		return;
	}
	composite_stay_quiet = 0;
	t_bug_on_2(
		!glyph_func,
		T_BUG("Glyph function is NULL"),
		return
	);

	(*glyph_func)(pThinc, pSrc, pDst, xSrc, ySrc, nlists, lists, glyphs); 

	thincDbg(T_DBG_DRAW, ("%s: finished\n", func));
}

static void thincGlyphs1Bpp(thincScreenPtr pThinc, PicturePtr pSrc, 
	PicturePtr pDst, INT16 xSrc, INT16 ySrc, 
	int nlists, GlyphListPtr lists, GlyphPtr *glyphs)
{
	t_dbg_name(func)
	BoxRec extents;
	int size, isOff, err;
	unsigned char *bitmap;
	RegionRec reg;
	unsigned long color;
	BoxPtr box;

	/* get the color first so we don't have to put the other stuff
	 * later */
	if(thincGetPictureColor(pThinc, pSrc, &color) == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't get color "
			"for GLYPH update\n", func));
		return;
	}
	thincDbg(T_DBG_DRAW, ("%s: Update's color is %#lx. src coords %d,%d\n",
		func, color, xSrc, ySrc));

	miGlyphExtents (nlists, lists, glyphs, &extents);
	print_box(T_DBG_DRAW, &extents, "glyph extents");

	if(box_is_empty(&extents)) {
		thincDbg(T_DBG_DRAW, ("%s: nothing to see here. move "
			"along. ok, ok, the box is empty, happy? "
			"finished!\n", func));
		return;
	}

#if 0	/* DEBUG */
	thincDbg(T_DBG_DRAW, ("%s: src alphaMap? %p, dst alphaMap? %p\n",
		func, pSrc->alphaMap, pDst->alphaMap));
	thincDbg(T_DBG_DRAW, ("%s: src repeat? %d. clientClipType? %d\n",
		func, pSrc->repeat, pSrc->clientClipType));
	really_print_reg(T_DBG_DRAW, pThinc->pScreen ,pDst->pCompositeClip, 
		"dst composite clip");
	thincDbg(T_DBG_DRAW, ("%s: dst x %d, dst y %d\n", func, 
		pDst->pDrawable->x, pDst->pDrawable->y));
#endif
	
	/* clip the update */
	REGION_INIT(pThinc->pScreen, &reg, &thincNullBox, 1);
	err = miComputeCompositeRegion (&reg, 
		pSrc, NULL /* pMask */, pDst, xSrc, ySrc,
		0, 0, /* mask coords */
		extents.x1+pDst->pDrawable->x, extents.y1+pDst->pDrawable->y, 
		box_width(&extents), box_height(&extents));
	if(!err) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't clip GLYPH "
			"update\n", func));
		return;
	}
	if(region_is_empty(&reg)) {
		thincDbg(T_DBG_DRAW, ("%s: Draw region is empty, finished\n",
			func));
		REGION_UNINIT(pThinc->pScreen, &reg);
		return;
	}

	/* TODO TODO TODO */
	if(REGION_NUM_RECTS(&reg) != 1) {
		T_BUG_ARG("Composite region has more than one rectangle (%ld) "
			"Implement me", REGION_NUM_RECTS(&reg));
		REGION_UNINIT(pThinc->pScreen, &reg);
		return;
	}

	/* the glyphs don't know anything about the drawables position so
	 * we need to undo that translate from miComputeCompositeRegion()
	 * momentarily */
	REGION_TRANSLATE(pThinc->pScreen, &reg, -pDst->pDrawable->x, 
		-pDst->pDrawable->y);
	box = REGION_EXTENTS(pThinc->pScreen, &reg);

	/* get the bitmap */
	size = thincGetBitmapFromGlyph(pThinc, lists, nlists, glyphs, 
		&bitmap, box);
	if(size == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't get bitmap "
			"for glyphs\n", func));
		REGION_UNINIT(pThinc->pScreen, &reg);
		return;
	} else if(size == 0) {
		thincDbg(T_DBG_DRAW, ("%s: Bitmap is empty. finished\n",
			func));
		REGION_UNINIT(pThinc->pScreen, &reg);
		return;
	}
	thincDbg(T_DBG_DRAW, ("%s: got bitmap of size %d\n", func, size));

	/* translate to dst coords before queueing */
	REGION_TRANSLATE(pThinc->pScreen, &reg, pDst->pDrawable->x, 
		pDst->pDrawable->y);
	box = REGION_EXTENTS(pThinc->pScreen, &reg);

	thincDbg(T_DBG_DRAW, ("%s: queueing GLYPH command\n", func));
	isOff = is_offscreen(pDst->pDrawable);
	if(isOff) {
		err = thincQueueCommand(pThinc, (PixmapPtr)pDst->pDrawable, 
			T_FB_GLYPH, &reg, bitmap, box->x1, box->y1,
			box_width(box), box_height(box), color, T_NO_FLAGS);
	} else {
		err = thincBufferNewCmdReg(pThinc, 
			(WindowPtr)pDst->pDrawable, T_FB_GLYPH, &reg, 
			bitmap, box->x1, box->y1, box_width(box), 
			box_height(box), color, T_NO_FLAGS);
	}

	if(err == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't buffer or "
			"queue new GLYPH command\n", func));
		goto out;
	}

	/* tis done */
	thincDbg(T_DBG_DRAW, ("%s: finished\n", func));
out:
	REGION_UNINIT(pThinc->pScreen, &reg);
	free(bitmap);
}

/* Finds the color that was used to fill pic, and returns T_OK or T_ERROR
 * To find the color, we need to do some sketchy stuff. In RENDER, the 
 * actual picture carries this information. In the case of 1bpp, we know
 * that:
 * - pSrc should be 1x1
 * - It's only pixel has the actual color.
 * So we have two options. We can either dig up the color from the associated
 * pixmap, or we can look for a solid fill command in there (there should 
 * only be one) and get the info from it.
 * We take the first option for now, since i'm in a hurry. Second option
 * would probably be cleaner.
 */
static int thincGetPictureColor(thincScreenPtr pThinc,
	PicturePtr pic, unsigned long *color)
{
	PixmapPtr pix;
	unsigned char *data;
#ifdef T_STRICT_CHECKS
	int isOff;
#endif

#ifdef T_STRICT_CHECKS
	isOff = is_offscreen(pic->pDrawable);
	if(!isOff) {
		T_BUG("Picture for Glyph color is not associated with a "
			"pixmap. How can I find the color?");
		return T_ERROR;
	}
	/* check dimensions. if it's not 1x1 something's going on */
	if(pic->pDrawable->width != 1 || pic->pDrawable->height != 1) {
		T_BUG_ARG_2("Picture's dimensions are not 1x1: %dx%d",
			pic->pDrawable->width, pic->pDrawable->height);
		return T_ERROR;
	}
#endif

	pix = (PixmapPtr)pic->pDrawable;
	data = thinc_getPixmap(pix);

	*color = *(unsigned long*)data;

	return T_OK;
}

/* macros to make the code below more pleasant to the eye. we don't use 
 * functions because this function needs to be wicked fast. apologies in
 * advance */
/* get the coordinates of the glyph */
#define get_glyph_coords(xg, yg, x, y, glyph) 	\
	(xg) = (x) - (glyph)->info.x;			\
	(yg) = (y) - (glyph)->info.y;

#define get_glyph_dim(glyph_width, glyph_height, glyph) \
	(glyph_width) = (glyph)->info.width;		\
	(glyph_height) = (glyph)->info.height;

#define next_glyph(x, y, glyph, glyphs)	\
	(x) += (glyph)->info.xOff;	\
	(y) += (glyph)->info.yOff;

/* we don't use bitmap_width cuz we want the floor, not the ceiling */
#define move_bitmap_pen(xpos, xpos_b, xpos_base, curr, bitmap, xOff) \
	(xpos) += (xOff);				\
	(xpos_b) = (xpos) >> T_LOG2_PIXELS_PER_BYTE;		\
	(xpos_base) = (xpos_b) << T_LOG2_PIXELS_PER_BYTE;		\
	(curr) = (bitmap) + (xpos_b);

/* true if glyph has gone over the right border of extents */
#define glyph_is_last(xg, extents)	((xg) >= (extents)->x2)

#define glyph_too_fat(xg, glyph_width, extents) \
	( (xg) + (glyph_width) > (extents)->x2)

#define glyph_too_high(yg, extents) ((yg) < (extents)->y1)

#define glyph_too_low(yg, extents) ((yg) >= (extents)->y2)

#define glyph_too_tall(yg, glyph_height, bitmap_y, bitmap_height) 	\
	( ( (glyph_height) > (bitmap_height) ) 		||		\
	  ( ((yg)+(glyph_height)) > ((bitmap_y)+(bitmap_height)) )	\
	)

/***********************************************/
#define glyph_to_box(box, glyph, x, y)			\
	(box)->x1 = (x) - (glyph)->info.x;		\
	(box)->y1 = (y) - (glyph)->info.y;		\
	(box)->x2 = (box)->x1 + (glyph)->info.width;	\
	(box)->y2 = (box)->y1 + (glyph)->info.height;

/* clip glyph to fit into bitmap */
#define glyph_clip(glyph, bitmap)				\
	(glyph)->x1 = MAX((glyph)->x1, (bitmap)->x1);		\
	(glyph)->x2 = MIN((glyph)->x2, (bitmap)->x2);		\
	(glyph)->x1 = MAX((glyph)->x1, (bitmap)->x1);		\
	(glyph)->y2 = MIN((glyph)->y2, (bitmap)->y2);

#define glyph_coords(xg, yg, glyph_box)			\
	(xg) = (glyph_box)->x1;				\
	(yg) = (glyph_box)->y1;

#define glyph_dimensions(g_widht, g_height, glyph_box)		\
	(g_width) = box_width_careful(glyph_box);		\
	(g_height) = box_height_careful(glyph_box);
	
/* returns size of bitmap or T_ERROR. it will also allocate memory for
 * the bitmap in bitmap and put it there. And the dimensions and position
 * of the bitmap will be returned in glyph_box */
int thincGetBitmapFromGlyph(thincScreenPtr pThinc, GlyphListPtr lists, 
	int nlists, GlyphPtr *glyphs, unsigned char **bitmap_p, 
	BoxPtr extents)
{
	t_dbg_name(func)
	int x, y, n, w = 0, h = 0, size, bw;
	int xg, yg;	/* where the glyph is actually drawn */
	int i, gwb;
	GlyphPtr glyph;
	unsigned char *bitmap, *curr, *linep, *glyph_data;
	int xpos, xpos_b, xpos_base, glyph_width, glyph_height;
	unsigned long bits;
	BoxRec orig;

	thincDbg(T_DBG_DRAW, ("%s\n", func));

	print_box(T_DBG_DRAW, extents, "glyphs extents");
	w = box_width(extents);
	h = box_height(extents);
	thincDbg(T_DBG_DRAW, ("%s: width %d height %d\n", func,
				w, h));

	x = -extents->x1;
	y = -extents->y1;

	copy_boxp(&orig, extents);
	box_translate(extents, x, y);

	print_box(T_DBG_DRAW, extents, "after translate");

	bw = bitmap_width(w);
	size = bw*h;
	thincDbg(T_DBG_DRAW, ("%s: bitmap is of byte width %d and size %d "
		"padded %d)\n", func, bw, size, size+3));

	/* calloc cuz we need to be zeroed out. we add 3 bytes so a
	 * bit_insert/extract at the end doesn't screw up memory after
	 * the bitmap */
	bitmap = calloc(size+3, 1);
	if(!bitmap) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't allocate "
			"memory for bitmap\n", func));
		return T_ERROR;
	}

	/* the rendering process goes like this (see miGlyphs in 
	 * render/miglyph.c)
	 * - we first need to build the mask. this is done by Compositing
	 *   the glyph (glyph data is at glyph+1) into a temporary mask
	 *   picture. this is done by looping through the lists and the
	 *   glyphs in the lists.
	 * - when we're finished building the mask, we composite the 
	 *   src with the newly built mask into the destination picture.
	 */
	thincDbg(T_DBG_DRAW, ("%s: starting draw at %d,%d. have %d lists of "
		"glyphs\n", func, x, y, nlists));

	/* 
	 * <apologies for the long and stupid comment but i'm sure 3 days from
	 * now i'll forget how this works. and i need stupid explanations>
	 *
	 * so the way we have to do things is somewhat limited by the
	 * bit extraction code (which cannot deal with offsets larger than
	 * 32).
	 * due to this limitation we need to keep moving our output position
	 * in the bitmap. so basically in xpos we keep track of our current 
	 * position in terms of pixels (which in this case, remember each pixel 
	 * is 1 bit). then, xpos_b has the byte offset from the start of the
	 * bitmap (basically on which byte/column of the bitmap we're currently
	 * adding stuff to). Finally, xpos_base is xpos_b converted back to
	 * bits. this is how much we need to shave off the glyph coordinates
	 * according to the position of our output pointer. an example:
	 * - xpos is currently 14 (we're drawing at coordinates (14,y)).
	 * - then xpos_b is going to be 1, and that's the column we're
	 *   using as our leftmost point. and xpos_base will be 8.
	 * - since our current left axis (or x=0 point) is at 8 (xpos_base)
	 *   then the coordinate where we need to draw is not (14,y) but 
	 *   (6,y).
	 * makes sense?
	 */
	xpos = 0;
	xpos_b = 0;
	xpos_base = 0;
	curr = bitmap;
	/* important: there's no implied ordering (in coordinates) across
	 * lists. i.e. just because one list draws to the right of the
	 * bitmap box doesn't mean that the next will do the same too */

	/* FIXME: The way we're cliping the glyphs to the box is nasty.
	 * we should actually do it in three steps:
	 * 1. find the glyph that's not to the left of the box. here we
	 * just advance our x and y (see goto next).
	 * 2. draw glyphs normally until the one that streches beyond the
	 * boxes' right border. here even on skip we advance everything
	 * (see goto next_advance)
	 * 3. skip over the rest of the glyphs. should we check them? i
	 * think they are ordered so maybe there's no need to.
	 *
	 * Finally, since it seems that there's no implied ordering among
	 * lists of glyphs, we may actually need to reset our pointer and
	 * position variables back to the beginning before processing
	 * each list. CHECK THIS.
	 */
	while (nlists--) {
		thincDbg(T_DBG_DRAW, ("%s: list information: "
			"format %#lx, depth %d\n", func, lists->format->format,
			lists->format->depth));

		thincDbg(T_DBG_DRAW, ("%s: list xoff %d yoff %d\n", func,
			lists->xOff, lists->yOff));
		x += lists->xOff;
		y += lists->yOff;
		n = lists->len;
		thincDbg(T_DBG_DRAW, ("%s: now drawing at %d %d. drawing "
			"%d glyphs\n", func, x, y, n));
		while(n--) {
			t_bug_on_2(
				curr < bitmap, 
				T_BUG_ARG_2("curr %p is before bitmap %p",
					curr, bitmap),
				T_BUG_ARG_2("pen is drawing before the bitmap "
					"memory! xpos_bytes %d, xpos %d\n",
					xpos_b, xpos)
			);
			glyph = *glyphs++;
			get_glyph_coords(xg, yg, x, y, glyph);
			get_glyph_dim(glyph_width, glyph_height, glyph);

			/* find first glyph */
			if(xg+glyph_width < extents->x1) {
				x += glyph->info.xOff;
				y += glyph->info.yOff;
				continue;
			}
			/* got first glyph */
			thincDbg(T_DBG_DRAW, ("%s: glyph info: x %d y %d "
				"width %d height %d\n", func, xg, yg,
				glyph_width, glyph_height));

			/* some glyphs have no width or height. although
			 * we don't draw them, they still contribute to
			 * the overall positioning of the glyph list */
			if(!glyph_width || !glyph_height)
				goto next_pen;

			/* clip the glyph as needed */
			/***** check x *****/
			if(glyph_is_last(xg, extents)) {
				thincDbg(T_DBG_DRAW, ("%s: glyph is beyond "
					"(%d) the bitmap's right border (%d). "
					"we're done with this list\n", func,
					xg, extents->x2));
				break;
			}
			if(glyph_too_fat(xg, glyph_width, extents)) {
				thincDbg(T_DBG_DRAW, ("%s: glyph draws "
					"outside outside of the box (%d), "
					"fixing its width to %d\n", func,
					xg+glyph_width,
					extents->x2 - xg));
				glyph_width = extents->x2 - xg;
			}

			/* check if as a result of the checking the glyph
			 * became invalid */
			if(glyph_width <= 0) {
				thincDbg(T_DBG_DRAW, ("%s: glyph's width "
					"%d not good, skipping\n", func,
					glyph_width));
				goto next;
			}

			/*** check y */
			if(glyph_too_high(yg, extents)) {
				thincDbg(T_DBG_DRAW, ("%s: glyph starts "
					"on top (%d) of bitmap box (%d) "
					"fixing it\n", func, yg,
					extents->y1));
				glyph_height -= extents->y1 - yg;
				yg = extents->y1; 
				if(glyph_height <= 0) goto next;
			}
			if(glyph_too_low(yg, extents)) {
				thincDbg(T_DBG_DRAW, ("%s: glyph starts "
					"below the box (%d vs %d) "
					"skipping\n", func, yg, 
					extents->y2));
				goto next;
			}
			if(glyph_too_tall(yg,glyph_height,extents->y1,h))
			{
				thincDbg(T_DBG_DRAW, ("%s: glyph "
					"draws outside of box (%d), fixing "
					"its height to %d\n", func,
					yg + glyph_height,
					extents->y2 - yg));
				glyph_height = (extents->y1+h) - yg;
			}

			thincDbg(T_DBG_DRAW, ("%s: after checking, glyph "
				"is at %d %d and dim %d %d\n", func, xg, yg,
				glyph_width, glyph_height));

			t_bug_on(
				(glyph_width > w),
				T_BUG_ARG_4("Glyph is wider (%d) than "
					"bitmap %d xg %d, extents->x2 %d", 
					glyph_width, w, xg, extents->x2)
			)
			t_bug_on(
				(glyph_height > h),
				T_BUG_ARG_4("Glyph is taller (%d) than "
					"bitmap %d, yg %d, bitmap_y %d", 
					glyph_height, h , yg, extents->y1)
			)

			/* the width of the glyph in bytes */
			gwb = bitmap_width_pad(glyph_width);
			glyph_data = (unsigned char*) (glyph+1);
#ifdef WRITE_GLYPH_TOFILE 
			/* write glyph to file */
			thincDumpGlyph(glyph_data, glyph_width,
				glyph_height);
#endif
			xg -= xpos_base;
			if(xg < 0) {
				thincDbg(T_DBG_DRAW, ("%s: had to backtrack\n",
					func));
				if(xpos_b) {
					/* need to backtrack */
					t_bug_on(
						(curr == bitmap),
						T_BUG_ARG_2("made curr invalid "
							"while backtracking. "
							"xpos %d, xg %d",
							xpos, xg)
					);
					curr--;
					xg+= T_PIXELS_PER_BYTE;
				} else {
					glyph_width += xg;
					xg = 0;
					thincDbg(T_DBG_DRAW, ("%s: was supposed "
						"to backtrack but nowhere "
						"to go, so clipping glyph. "
						"xg %d, width %d\n", func,
						xg, glyph_width));

				}
			} 
			thincDbg(T_DBG_DRAW, ("%s: n %d curr %p, bitmap %p, "
				"xpos %d, xpos_b %d\n", func, n, curr, 
				bitmap, xpos, xpos_b));
			t_bug_on(
				(curr < bitmap),
				T_BUG_ARG_2("curr %p is to the left of the "
					"bitmap %p", curr, bitmap)
			)
			t_bug_on(
				(curr - bitmap >= bw),
				T_BUG_ARG_2("curr has gone over the bitmap "
					"right border. We are %d bytes "
					"away from the left border, but "
					"width is %d", curr-bitmap, w)
			);
			t_bug_on(
				((curr + bitmap_width(glyph_width)) > bitmap+bw),
				T_BUG_ARG_2("we just blew our memory, %p %p",
					curr+gwb, bitmap+bw)
			);
			/* go to line where glyph starts */
			linep = curr + (bw*yg);
			for(i=0; i<glyph_height; ++i) {
				int off = MAX(0, glyph->info.x);
				thincDbg(T_DBG_DRAW, ("%s: bitmap offset %d\n", 
					func, off));
				bits = t_bit_extract(glyph_data, off, 
					glyph_width);
				t_bug_on(
					(linep < curr),
					T_BUG_ARG_2("drawing glyph to "
						"memory before the bitmap! "
						"linep %p, bw %d", linep,
						bw)
				);
				t_bug_on(
					(linep < bitmap),
					T_BUG_ARG_2("curr %p is before "
						"bitmap %p", curr,
						bitmap)
				);
				t_bug_on_2(
					(linep+1 > bitmap+size),
					T_BUG_ARG_4("inserting outside of "
						"bitmap, i is %d, "
						"glyph_height %d bitmap height "
						"%d curr %p",
						i, glyph_height, h, curr),
					T_BUG_ARG_3("bw is %d, gwb %d yg %d", 
						bw, gwb, yg)
				);
				t_bit_insert(linep, xg, glyph_width, bits);
				linep += bw;
				glyph_data += gwb;
			}
#ifdef WRITE_GLYPH_TOFILE /* write current version of bitmap to file */
			thincDumpGlyph(bitmap, w, h);
#endif

next_pen:
			thincDbg(T_DBG_DRAW, ("%s: advancing pen, xoff %d "
				"yoff %d\n", func, glyph->info.xOff,
				glyph->info.yOff));
			move_bitmap_pen(xpos, xpos_b, xpos_base, curr, bitmap,
				glyph->info.xOff);
			t_bug_on(
				(curr < bitmap),
				T_BUG_ARG_2("pen is drawing before the bitmap "
					"memory! xpos_bytes %d, xpos %d\n",
					xpos_b, xpos)
			);
			thincDbg(T_DBG_DRAW, ("%s: moved pen. n %d curr %p, "
				"bitmap %p, xpos %d, xpos_b %d\n", func, n, 
				curr, bitmap, xpos, xpos_b));

next:
			next_glyph(x, y, glyph, glyphs);
		}
		++lists;
#if 1		/* XXX HACK HACK HACK XXX */
		if(lists->xOff > 0) {
			move_bitmap_pen(xpos, xpos_b, xpos_base, curr, bitmap,
				lists->xOff);
		}
#endif
	}

	copy_boxp(extents, &orig);

	*bitmap_p = bitmap;

	thincDbg(T_DBG_DRAW, ("%s: finished. bitmap size is %d\n", func, 
		size));
	return size;
}

void thincTrapezoids(CARD8 op, PicturePtr pSrc, PicturePtr pDst,
	PictFormatPtr maskFormat, INT16 xSrc, INT16 ySrc, int ntrap, 
	xTrapezoid *traps)
{
	t_dbg_name(func)
	thincPict_pre(pSrc, Trapezoids);

	thincDbg(T_DBG_DRAW, ("%s: op %d\n", func, op));
	thincDrawable_print(pSrc->pDrawable, "src: ");
	thincDrawable_print(pDst->pDrawable, "dst: ");

	ps->Trapezoids(op, pSrc, pDst, maskFormat, xSrc, ySrc, ntrap,
		traps);
	thincPict_post(ps, pThinc, Trapezoids);

	thincDbg(T_DBG_DRAW, ("%s: finished\n", func));
}

void thincTriangles (CARD8 op, PicturePtr pSrc, PicturePtr pDst,
	PictFormatPtr maskFormat, INT16 xSrc, INT16 ySrc, int ntri, 
	xTriangle *tris)
{
	t_dbg_name(func)
	thincPict_pre(pSrc, Triangles);

	thincDbg(T_DBG_DRAW, ("%s: op %d\n", func, op));
	thincDrawable_print(pSrc->pDrawable, "src: ");
	thincDrawable_print(pDst->pDrawable, "dst: ");

	ps->Triangles(op, pSrc, pDst, maskFormat, xSrc, ySrc, ntri,
		tris);
	thincPict_post(ps, pThinc, Triangles);

	thincDbg(T_DBG_DRAW, ("%s: finished\n", func));
}

void thincTriStrip (CARD8 op, PicturePtr pSrc, PicturePtr pDst,
	PictFormatPtr maskFormat, INT16 xSrc, INT16 ySrc, int npoint, 
	xPointFixed *points)
{
	t_dbg_name(func)
	thincPict_pre(pSrc, TriStrip);

	thincDbg(T_DBG_DRAW, ("%s: op %d\n", func, op));
	thincDrawable_print(pSrc->pDrawable, "src: ");
	thincDrawable_print(pDst->pDrawable, "dst: ");

	ps->TriStrip(op, pSrc, pDst, maskFormat, xSrc, ySrc, npoint,
		points);
	thincPict_post(ps, pThinc, TriStrip);

	thincDbg(T_DBG_DRAW, ("%s: finished\n", func));
}

void thincTriFan (CARD8 op, PicturePtr pSrc, PicturePtr pDst,
	PictFormatPtr maskFormat, INT16 xSrc, INT16 ySrc, int npoint, 
	xPointFixed *points)
{
	t_dbg_name(func)
	thincPict_pre(pSrc, TriFan);

	thincDbg(T_DBG_DRAW, ("%s: op %d\n", func, op));
	thincDrawable_print(pSrc->pDrawable, "src: ");
	thincDrawable_print(pDst->pDrawable, "dst: ");

	ps->TriFan(op, pSrc, pDst, maskFormat, xSrc, ySrc, npoint,
		points);
	thincPict_post(ps, pThinc, TriFan);

	thincDbg(T_DBG_DRAW, ("%s: finished\n", func));
}

void thincRasterizeTrapezoid (PicturePtr pMask, xTrapezoid *trap,
	int x_off, int y_off)
{
	t_dbg_name(func)
	thincPict_pre(pMask, RasterizeTrapezoid);

	thincDbg(T_DBG_DRAW, ("%s\n", func));
	thincDrawable_print(pMask->pDrawable, "");

	ps->RasterizeTrapezoid(pMask, trap, x_off, y_off);
	thincPict_post(ps, pThinc, RasterizeTrapezoid);

	thincDbg(T_DBG_DRAW, ("%s: finished\n", func));
}

#ifdef WRITE_GLYPH_TOFILE_FUN
/* Assumes bitmap is padded so if printing our bitmaps valgrind may complain.
 * but that's ok, we're only reading bogus data */
static void thincDumpGlyph(unsigned char *data, int width, int height)
{
	static int count = 0;
	const char *base = "/tmp/glyph/glyph";
	char fn[255];
	const int mode = S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH;
	int dump, write_header = 1;
	unsigned int a, size, w2;

	strncpy(fn, base, 16);
	sprintf(fn+16, "%d", count++);
	thincError("writing file %s\n", fn);

	dump = creat(fn,mode);

	/* X bitmaps are padded to 4 bytes */
	w2 = bitmap_width_pad(width);
	size = w2*height;

	if(write_header) {
		a = htonl(width);
		write(dump, &a, 4);
		a = htonl(height);
		write(dump, &a, 4);
		a = htonl(size);
		write(dump, &a, 4);
		a = 0;
		write(dump, &a, 4);
	}
	write(dump, data, size);
	close(dump);
}
#endif	/* WRITE_GLYPH_TOFILE_FUN */

#endif /* T_USERENDER */
