/* Copyright 2003, 2004, 2005 Ricardo Baratto */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/

/*
 * $Id; $
 *
 * SSL client functions
 *
 * Author:
 * Angelos Stavrou <angel@cs.columbia.edu>
 * Stelianos Sidiroglou-Douskos <stelios@cs.columbia.edu>
 * Network Security Lab - Columbia University
 *
 * Modified for THINC:
 * Ricardo Baratto
 * Network Computing Lab - Columbia University
 */

/* set tabstop = 8 */

/*
 * to talk:
 * thinc_ssl_connect()
 *
 * then:
 * SSL_write(fd)
 * SSL_read(fd)
 *
 * TODO: Wrap SSL_write() and SSL_read()
 */

#include "thinc_ssl.h"

static SSL_CTX* thinc_ssl_init_ctx(const char * cert_location);

/*
 * Opens an SSL connection using the socket sock
 * Returns 1 on success, -1 on error.
 * May not work with non-blocking sockets. sorry, i'm lazy :)
 */
int thinc_ssl_connect(int sock, const char *cert_location, SSL **ssl_p)
{
	SSL_CTX *ctx;
	SSL *ssl;
	int err, ssl_err;

	if(*ssl_p) {
		ssl = *ssl_p;
	} else {
		ctx = thinc_ssl_init_ctx(cert_location);
		/* create new SSL connection state */
		ssl = SSL_new(ctx);		

		/* use stream cipher */
		SSL_set_cipher_list(ssl, SSL_TXT_RC4);
		/* attach the socket descriptor */
		SSL_set_fd(ssl, sock);
	}
	err = SSL_connect(ssl);
	if(err == 0 ) {
		/* connection closed */
		SSL_free(ssl);
		*ssl_p = NULL;
		return T_SSL_CLOSED;
	} else if(err < 0) {
		/* check what kind of error */
		ssl_err = SSL_get_error(ssl, err);
		if( 	(ssl_err == SSL_ERROR_WANT_READ) || 
			(ssl_err == SSL_ERROR_WANT_WRITE) ) {
			/* non-blocking BIO, tell caller */
			*ssl_p = ssl;
			return T_SSL_WOULDBLOCK;
		} else {
			SSL_free(ssl);
			*ssl_p = NULL;
			return T_SSL_ERROR;
		}
	}
	*ssl_p = ssl;
	return T_SSL_OK;
}

int thinc_ssl_read(SSL *ssl, void *buf, int count)
{
	int err, ssl_err;

again:
	err = SSL_read(ssl, buf, count);
	if(err == 0) {
		return T_SSL_CLOSED;
	} else if(err < 0) {
		ssl_err = SSL_get_error(ssl, err);
		/* man page says both errors can be returned since at any
		 * point in time, a handshake may be necessary... */
		if( 	(ssl_err == SSL_ERROR_WANT_READ) || 
			(ssl_err == SSL_ERROR_WANT_WRITE) ) {
			return T_SSL_WOULDBLOCK;
		} else if( ssl_err == SSL_ERROR_SYSCALL) {
			goto again;
		} else {
			return T_SSL_ERROR;
		}
	}

	return err;
}

/*
 * Initializes SSL context
 */
static SSL_CTX* thinc_ssl_init_ctx(const char * cert_location)
{
	SSL_METHOD *method;
	SSL_CTX *ctx;

	OpenSSL_add_all_algorithms();	/* Load cryptos, et.al. */
	SSL_load_error_strings();	/* Bring in and register error 
					   messages */
	/* Create new client-method instance */
	method = SSLv23_method();
	ctx = SSL_CTX_new(method);	/* Create new context */
	if ( ctx == NULL )
		return NULL;

#if 0
	SSL_CTX_load_verify_locations(ctx, cert_location, 0);
	SSL_CTX_set_verify(ctx,SSL_VERIFY_PEER,NULL);
#endif
	return ctx;
}
