/* Copyright 2004, 2005 Lei Zhang, Columbia University */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/
/******************************Module*Header*******************************\
*
*                       **************************
*                       * THINC Related Routines *
*                       **************************
*
* Module Name: thincCmdBlock.c
*
* This module contains the functions that are related with THINC 
*
***************************************************************************/


#include "thincCmdBlock.h"
#include "driver.h"


PThincCmdBlock thincCmdBlockCreateMax()
{
	return thincCmdBlockCreate(MAX_CMD_BLOCK_SIZE);
}

VOID thincCmdBlockInit(PThincCmdBlock pCmdBlock)
{    
	//pCmdBlock->availableBytes = blockSize;
	pCmdBlock->writeOffset = 0;
	pCmdBlock->readOffset = 0;
	
	//pCmdBlock->bNoMoreRead = FALSE;
	//pCmdBlock->bNoMoreWrite = FALSE;
	//pCmdBlock->psurf = psurf;
	//pCmdBlock->pNextBlock_kernelVA = NULL;
	//pCmdBlock->pNextBlock_userVA = NULL;
	pCmdBlock->hMappingHandle = NULL;
}
PThincCmdBlock thincCmdBlockCreate(INT blockSize)
{
	PThincCmdBlock pCmdBlock = EngAllocMem(	FL_NONPAGED_MEMORY,
											sizeof(ThincCmdBlock) + blockSize,
											MKTAG("CBLK"));
	if (pCmdBlock == NULL)
	{
		return NULL;
	}
	else
	{
		pCmdBlock->blockSize = blockSize;
	}
	
	thincCmdBlockInit(pCmdBlock);

	return pCmdBlock;
}

PVOID thincCmdBlockAlloc(PThincCmdBlock* ppThincCmdBlock, INT size, INT *pOffset)
{                             
	PThincCmdBlock	pThincCmdBlock;
	PVOID			pCmd;
	INT				iCmdBlkSize;
	INT				availableBytes;
	INT				thincCmdSize;

	ASSERT((ppThincCmdBlock != NULL), (" thincCmdBlockAlloc: pThincCmdBlock == NULL\n"));
	ASSERT((*ppThincCmdBlock != NULL), (" thincCmdBlockAlloc: *pThincCmdBlock == NULL\n"));
	
	pThincCmdBlock = *ppThincCmdBlock;

	thincCmdSize = size + sizeof(ThincCmd);

	if (thincCmdSize & 0x03)
	{
		thincCmdSize += 4 - thincCmdSize & 0x03;
	}

	if (pThincCmdBlock->readOffset > pThincCmdBlock->writeOffset)
	{
		availableBytes = pThincCmdBlock->readOffset	- pThincCmdBlock->writeOffset;
	}
	else
	{
		availableBytes = pThincCmdBlock->blockSize - pThincCmdBlock->writeOffset;
	}
                                        
	if (thincCmdSize > availableBytes)
	{
		//
		// FIXME: We assume that the most commands in the beginning of the block has been read
		//
		pThincCmdBlock->readOffset = pThincCmdBlock->writeOffset;

		if (thincCmdSize < pThincCmdBlock->readOffset)
		{
			//
			// [0..readOffset] can be reused.
			//
			pThincCmdBlock->writeOffset = 0;
		}
		else
		{
			//
			// No enough space in current block to go around 
			// 
			RIP("Not enough space in the command block\n");
		}

	}

	//
	// Allocate a chunck for the new command
	// 
	pCmd = (PCHAR)pThincCmdBlock + sizeof(ThincCmdBlock) + pThincCmdBlock->writeOffset;
	((PThincCmd)(pCmd))->size = size;
	
	//
	// Copy the command block pointer and write offset to the command queue;
	//

	//pDestThincCmdBlock->availableBytes -= size + sizeof(ThincCmd);
	*pOffset = sizeof(ThincCmdBlock) + pThincCmdBlock->writeOffset;
	pThincCmdBlock->writeOffset += thincCmdSize;

	return pCmd;
}



VOID thincCmdBlockDestroy(PThincCmdBlock pThincCmdBlock)
{
	ASSERT((pThincCmdBlock != NULL), (" thincCmdBlockDestroy: pThincCmdBlock == NULL\n"));
	//
	// Unmap the command block
	//
	if (pThincCmdBlock->hMappingHandle != NULL)
	{
		//pThincCmdBlock->psurf->ppdev->pfUnMapMem(pThincCmdBlock->hMappingHandle);
	}
	EngFreeMem(pThincCmdBlock);
}

PThincCmdBlock thincCmdBlockDestroyList(PThincCmdBlock pThincCmdBlock)
{             
	PThincCmdBlock pPrevThincCmdBlock;
	PPDEV ppdev;

	//while (pThincCmdBlock != NULL && pThincCmdBlock->bNoMoreWrite && pThincCmdBlock->bNoMoreRead)
	//{
	//	pPrevThincCmdBlock = pThincCmdBlock;
	//	//pThincCmdBlock = pThincCmdBlock->pNextBlock_kernelVA;
	//	thincCmdBlockDestroy(pPrevThincCmdBlock);
 // 	}
	return pThincCmdBlock;
}
