/* Copyright 2004, 2005 Lei Zhang, Columbia University */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/
/******************************Module*Header*******************************\
*
*                           *********************
*                           * THINC header file *
*                           *********************
*
* Module Name: thincCmdBlock.h
*
* This module contains the functions that are related with THINC 
*
\**************************************************************************/
#ifndef _THINC_CMD_LIST_
#define _THINC_CMD_LIST_

#include "driver.h"

#define MAX_CMD_BLOCK_SIZE		0x1000000	// 16MB
#define NORMAL_CMD_BLOCK_SIZE	0x10000		// 64KB

typedef struct ThincCmd_s {
	INT				size;
	char			timeStamp[16];
	char			cmdData[0];
	// followed by size of bytes
} ThincCmd, *PThincCmd;

typedef struct ThincCmdBlock_s {
	//struct ThincCmdBlock_s* pNextBlock_kernelVA;
	//struct ThincCmdBlock_s* pNextBlock_userVA;
	//PMirrSurf		psurf;				// the dest surf of these commands
	HANDLE			hMappingHandle;		// mapping handle for this block
	PVOID			userVirtualAddress;	// mapped user mode address of this block
	INT				blockSize;			// size of the block, excluding the struct size
	//INT				availableBytes;		// available bytes for new commands
	INT				writeOffset;		// offset of available bytes from the first byte of this struct
	INT				readOffset;			// offset of reading
	//BOOL			bNoMoreRead;		// no other commands will be writen to this block
	//BOOL			bNoMoreWrite;		// the user mode application has read the whole block.
	//									// this flag will not be set until the bNotMoreWrite is set.
	//BOOL			bNewCmds;			// indicates whether there are new commands in this block
	//PThincDrvNode	pNode;
	// followed by blockSize of bytes
} ThincCmdBlock, *PThincCmdBlock;

// Initialize the command block
VOID thincCmdBlockInit(PThincCmdBlock pCmdBlock);

// Create and initialize a command block with maximal size
PThincCmdBlock thincCmdBlockCreateMax();

// Create and initialize a command block with specified size
PThincCmdBlock thincCmdBlockCreate(INT blockSize);

// Alocate size of bytes in memory block to store a memory. Update pointer in 
// ppThincCmdBlock for the caller surface when necessary. Return the memory block
// that has been allocated, return NULL when allocating fails
PVOID thincCmdBlockAlloc(PThincCmdBlock* ppThincCmdBlock, INT size, INT *pOffset);

// thincCmdBlockAllocCmd and thincCmdBlockCommitCmd should be paired up.
// this function must be called after the command has been writen to the block.
PThincCmdBlock thincCmdBlockCommitCmd(PThincCmdBlock pThincCmdBlock, INT size);

// Destroy a command block
VOID thincCmdBlockDestroy(PThincCmdBlock pThincCmdBlock);

// Destroy the command blocks that are not in used.
// Return the next command block in the list
PThincCmdBlock thincCmdBlockDestroyList(PThincCmdBlock pThincCmdBlock);

#endif /* _THINC_CMD_LIST_ */
